from typing import Optional, FrozenSet, Mapping

from immutables import Map

from HABApp.openhab.items.base_item import OpenhabItem, MetaData
from HABApp.openhab.items.commands import OnOffCommand, PercentCommand
from ..definitions import OnOffValue, PercentValue


class DimmerItem(OpenhabItem, OnOffCommand, PercentCommand):
    """DimmerItem which accepts and converts the data types from OpenHAB

    :ivar str name:
    :ivar Union[int, float] value:

    :ivar Optional[str] label:
    :ivar FrozenSet[str] tags:
    :ivar FrozenSet[str] groups:
    :ivar Mapping[str, MetaData] metadata:
    """

    @classmethod
    def from_oh(cls, name: str, value=None, label: Optional[str] = None, tags: FrozenSet[str] = frozenset(),
                groups: FrozenSet[str] = frozenset(), metadata: Mapping[str, MetaData] = Map()):
        if value is not None:
            value = float(value)
        return cls(name, value, label=label, tags=tags, groups=groups, metadata=metadata)

    def set_value(self, new_value) -> bool:

        if isinstance(new_value, OnOffValue):
            new_value = 100 if new_value.on else 0
        elif isinstance(new_value, PercentValue):
            new_value = new_value.value

        # Percent is 0 ... 100
        if isinstance(new_value, (int, float)):
            assert 0 <= new_value <= 100, new_value
        else:
            assert new_value is None, new_value

        return super().set_value(new_value)

    def __str__(self):
        return self.value

    def is_on(self) -> bool:
        """Test value against on-value"""
        return bool(self.value)

    def is_off(self) -> bool:
        """Test value against off-value"""
        return not bool(self.value)



class ItemRegistryItem:
    """ItemRegistryItem, all items that will be stored in the Item Registry must inherit from this
    """

    def __init__(self, name: str, **kwargs):
        super().__init__(**kwargs)
        assert isinstance(name, str), type(name)
        self._name: str = name

    @property
    def name(self) -> str:
        """
        :return: Name of the item (read only)
        """
        return self._name

    def _on_item_added(self):
        """This function gets automatically called when the item was added to the item registry
        """
        raise NotImplementedError()

    def _on_item_removed(self):
        """This function gets automatically called when the item was removed from the item registry
        """
        raise NotImplementedError()



import logging
from typing import Optional, Callable

from HABApp.core.asyncio import async_context, create_task
from HABApp.core.internals import HINT_CONTEXT_OBJ
from .base import WrappedFunctionBase


class WrappedSyncFunction(WrappedFunctionBase):

    def __init__(self, func: Callable,
                 warn_too_long=True,
                 name: Optional[str] = None,
                 logger: Optional[logging.Logger] = None,
                 context: Optional[HINT_CONTEXT_OBJ] = None):

        super().__init__(name=name, func=func, logger=logger, context=context)
        assert callable(func)

        self.func = func
        self.warn_too_long: bool = warn_too_long

    def run(self, *args, **kwargs):
        create_task(self.async_run(*args, **kwargs), name=self.name)

    async def async_run(self, *args, **kwargs):

        token = async_context.set('WrappedSyncFunction')

        try:
            self.func(*args, **kwargs)
        except Exception as e:
            self.process_exception(e, *args, **kwargs)
        finally:
            async_context.reset(token)



from typing import Iterable as _Iterable


class CircularReferenceError(Exception):
    def __init__(self, stack: _Iterable[str]):
        self.stack = stack

    def __repr__(self):
        return f'<{self.__class__.__name__} {" -> ".join(self.stack)}>'


class DependencyDoesNotExistError(Exception):
    def __init__(self, msg: str):
        self.msg = msg

    def __repr__(self):
        return f'<{self.__class__.__name__} {self.msg}>'


class AlreadyHandledFileError(Exception):
    pass



import asyncio
from pathlib import Path

import HABApp
import HABApp.config
import HABApp.core
import HABApp.mqtt.mqtt_connection
import HABApp.parameters.parameter_files
import HABApp.rule.interfaces._http
import HABApp.rule_manager
import HABApp.util
import eascheduler
from HABApp.core.asyncio import async_context
from HABApp.core.internals import setup_internals
from HABApp.core.internals.proxy import ConstProxyObj
from HABApp.core.wrapper import process_exception
from HABApp.openhab import connection_logic as openhab_connection
from HABApp.runtime import shutdown


class Runtime:

    def __init__(self):
        self.config: HABApp.config.Config = None

        # Rule engine
        self.rule_manager: HABApp.rule_manager.RuleManager = None

    async def start(self, config_folder: Path):
        try:
            token = async_context.set('HABApp startup')

            # setup exception handler for the scheduler
            eascheduler.set_exception_handler(lambda x: process_exception('HABApp.scheduler', x))

            # Start Folder watcher!
            HABApp.core.files.watcher.start()

            # Load config
            HABApp.config.load_config(config_folder)

            # replace proxy objects
            ir = HABApp.core.internals.ItemRegistry()
            eb = HABApp.core.internals.EventBus()
            setup_internals(ir, eb)
            assert isinstance(HABApp.core.Items, ConstProxyObj)
            HABApp.core.Items = ir
            assert isinstance(HABApp.core.EventBus, ConstProxyObj)
            HABApp.core.EventBus = eb

            await HABApp.core.files.setup()

            # generic HTTP
            await HABApp.rule.interfaces._http.create_client()

            # openhab
            openhab_connection.setup()

            # Parameter Files
            await HABApp.parameters.parameter_files.setup_param_files()

            # Rule engine
            self.rule_manager = HABApp.rule_manager.RuleManager(self)
            await self.rule_manager.setup()

            # MQTT
            HABApp.mqtt.mqtt_connection.setup()
            HABApp.mqtt.mqtt_connection.connect()

            await openhab_connection.start()

            async_context.reset(token)

        except HABApp.config.InvalidConfigError:
            shutdown.request_shutdown()
        except Exception as e:
            process_exception('Runtime.start', e)
            await asyncio.sleep(1)  # Sleep so we can do a graceful shutdown
            shutdown.request_shutdown()



import logging
import re
import warnings
from typing import Iterable, Union, Any, Optional, Tuple, Pattern, List

import HABApp
import HABApp.core
import HABApp.openhab
import HABApp.rule_manager
import HABApp.util
from HABApp.core.asyncio import create_task
from HABApp.core.const.hints import HINT_EVENT_CALLBACK
from HABApp.core.internals import HINT_EVENT_FILTER_OBJ, HINT_EVENT_BUS_LISTENER, ContextProvidingObj, \
    uses_post_event, EventFilterBase, uses_item_registry, ContextBoundEventBusListener
from HABApp.core.internals import wrap_func
from HABApp.core.items import BaseItem, HINT_ITEM_OBJ, HINT_TYPE_ITEM_OBJ, BaseValueItem
from HABApp.rule import interfaces
from HABApp.rule.scheduler import HABAppSchedulerView as _HABAppSchedulerView
from .interfaces import async_subprocess_exec
from .rule_hook import get_rule_hook as _get_rule_hook

log = logging.getLogger('HABApp.Rule')


# Func to log deprecation warnings
def send_warnings_to_log(message, category, filename, lineno, file=None, line=None):
    log.warning('%s:%s: %s:%s' % (filename, lineno, category.__name__, message))
    return


# Setup deprecation warnings
warnings.simplefilter('default')
warnings.showwarning = send_warnings_to_log


post_event = uses_post_event()
item_registry = uses_item_registry()


class Rule(ContextProvidingObj):
    def __init__(self):
        super().__init__(context=HABApp.rule_ctx.HABAppRuleContext(self))

        hook = _get_rule_hook()
        hook.register_rule(self)

        self.__runtime: HABApp.runtime.Runtime = hook.runtime
        assert isinstance(self.__runtime, HABApp.runtime.Runtime)

        # scheduler
        self.run: _HABAppSchedulerView = _HABAppSchedulerView(self._habapp_ctx)

        # suggest a rule name
        self.rule_name: str = hook.suggest_rule_name(self)

        # interfaces
        self.async_http = interfaces.http
        self.mqtt: HABApp.mqtt.interface = HABApp.mqtt.interface
        self.oh: HABApp.openhab.interface = HABApp.openhab.interface
        self.openhab: HABApp.openhab.interface = self.oh

    def on_rule_loaded(self):
        """Override this to implement logic that will be called when the rule and the file has been successfully loaded
        """

    def on_rule_removed(self):
        """Override this to implement logic that will be called when the rule has been unloaded.
        """

    def __repr__(self):
        # empty string, so we have a space if we have more than one entry
        parts = ['']

        # rule name if it is different from the class
        cls_name = self.__class__.__name__
        rule_name = str(self.rule_name)
        if cls_name != rule_name:
            parts.append(rule_name)

        # rule status
        if self._habapp_ctx is None:
            parts.append('(unloaded)')

        return f'<{cls_name}{" ".join(parts)}>'

    def post_event(self, name: Union[HINT_ITEM_OBJ, str], event: Any):
        """
        Post an event to the event bus

        :param name: name or item to post event to
        :param event: Event class to be used (must be class instance)
        :return:
        """
        assert isinstance(name, (str, BaseValueItem)), type(name)
        return post_event(
            name.name if isinstance(name, BaseValueItem) else name,
            event
        )

    def listen_event(self, name: Union[HINT_ITEM_OBJ, str],
                     callback: HINT_EVENT_CALLBACK,
                     event_filter: Optional[HINT_EVENT_FILTER_OBJ] = None
                     ) -> HINT_EVENT_BUS_LISTENER:
        """
        Register an event listener

        :param name: item or name to listen to
        :param callback: callback that accepts one parameter which will contain the event
        :param event_filter: Event filter. This is typically :class:`~HABApp.core.events.ValueUpdateEventFilter` or
            :class:`~HABApp.core.events.ValueChangeEventFilter` which will also trigger on changes/update from openhab
            or mqtt. Additionally it can be an instance of :class:`~HABApp.core.events.EventFilter` which additionally
            filters on the values of the event. It is also possible to group filters logically with, e.g.
            :class:`~HABApp.core.events.AndFilterGroup` and :class:`~HABApp.core.events.OrFilterGroup`
        """
        cb = wrap_func(callback, context=self._habapp_ctx)
        name = name.name if isinstance(name, BaseItem) else name

        if event_filter is None:
            event_filter = HABApp.core.events.NoEventFilter()
        if not isinstance(event_filter, EventFilterBase):
            raise ValueError(f'Argument event_filter must be an instance of event filter (is {event_filter})')

        listener = ContextBoundEventBusListener(name, cb, event_filter, parent_ctx=self._habapp_ctx)
        return self._habapp_ctx.add_event_listener(listener)

    def execute_subprocess(self, callback: HINT_EVENT_CALLBACK, program, *args, capture_output=True):
        """Run another program

        :param callback: |param_scheduled_cb| after process has finished. First parameter will
                         be an instance of :class:`~HABApp.rule.FinishedProcessInfo`
        :param program: program or path to program to run
        :param args: |param_scheduled_cb_args|
        :param capture_output: Capture program output, set to `False` to only capture return code
        :return:
        """

        assert isinstance(program, str), type(program)
        cb = wrap_func(callback, context=self._habapp_ctx)

        create_task(
            async_subprocess_exec(cb.run, program, *args, capture_output=capture_output),
        )

    def get_rule(self, rule_name: str) -> 'Union[Rule, List[Rule]]':
        assert rule_name is None or isinstance(rule_name, str), type(rule_name)
        return self.__runtime.rule_manager.get_rule(rule_name)

    @staticmethod
    def get_items(type: Union[Tuple[HINT_TYPE_ITEM_OBJ, ...], HINT_TYPE_ITEM_OBJ] = None,
                  name: Union[str, Pattern[str]] = None,
                  tags: Union[str, Iterable[str]] = None,
                  groups: Union[str, Iterable[str]] = None,
                  metadata: Union[str, Pattern[str]] = None,
                  metadata_value: Union[str, Pattern[str]] = None,
                  ) -> Union[List[HINT_ITEM_OBJ], List[BaseItem]]:
        """Search the HABApp item registry and return the found items.

        :param type: item has to be an instance of this class
        :param name: str (will be compiled) or regex that is used to search the Name
        :param tags: item must have these tags (will return only instances of OpenhabItem)
        :param groups: item must be a member of these groups (will return only instances of OpenhabItem)
        :param metadata: str (will be compiled) or regex that is used to search the metadata (e.g. 'homekit')
        :param metadata_value: str (will be compiled) or regex that is used to search the metadata value
                               (e.g. 'TargetTemperature')
        :return: Items that match all the passed criteria
        """

        if name is not None and isinstance(name, str):
            name = re.compile(name, re.IGNORECASE)
        if metadata is not None and isinstance(metadata, str):
            metadata = re.compile(metadata, re.IGNORECASE)
        if metadata_value is not None and isinstance(metadata_value, str):
            metadata_value = re.compile(metadata_value, re.IGNORECASE)

        _tags, _groups = None, None
        if tags is not None:
            _tags = set(tags) if not isinstance(tags, str) else {tags}
        if groups is not None:
            _groups = set(groups) if not isinstance(groups, str) else {groups}

        OpenhabItem = HABApp.openhab.items.OpenhabItem
        if _tags or _groups or metadata or metadata_value:
            if type is None:
                type = OpenhabItem
            if not issubclass(type, OpenhabItem):
                raise ValueError('Searching for tags, groups and metadata only works for OpenhabItem or its Subclasses')

        ret = []
        for item in item_registry.get_items():  # type: HABApp.core.items.BaseItem
            if type is not None and not isinstance(item, type):
                continue

            if name is not None and not name.search(item.name):
                continue

            if _tags is not None and not _tags.issubset(item.tags):
                continue

            if _groups is not None and not _groups.issubset(item.groups):
                continue

            if metadata is not None and not any(map(metadata.search, item.metadata)):
                continue

            if metadata_value is not None and not any(
                    map(metadata_value.search, map(lambda x: x[0], item.metadata.values()))):
                continue

            ret.append(item)
        return ret



from typing import Any

from HABApp.core.const import MISSING
from HABApp.core.events import ValueUpdateEvent, ValueChangeEvent
from HABApp.core.events.filter.event import TypeBoundEventFilter


class ValueUpdateEventFilter(TypeBoundEventFilter):
    def __init__(self, value: Any = MISSING):
        super().__init__(ValueUpdateEvent, value=value)


class ValueChangeEventFilter(TypeBoundEventFilter):
    def __init__(self, value: Any = MISSING, old_value: Any = MISSING):
        super().__init__(ValueChangeEvent, value=value, old_value=old_value)



import asyncio
import sys

# setup everything so we can create a subprocess, only required for older versions
if sys.version_info < (3, 8):
    if sys.platform == "win32":
        # This is the default from 3.8 so we don't have to set it ourselves
        asyncio.set_event_loop_policy(asyncio.WindowsProactorEventLoopPolicy())
    else:
        # Must be called once in the main thread so creating a subprocess works properly
        # https://docs.python.org/3/library/asyncio-subprocess.html#subprocess-and-threads
        asyncio.get_child_watcher()


loop = asyncio.get_event_loop_policy().get_event_loop()
loop.set_debug(True)
loop.slow_callback_duration = 0.02



from typing import Optional, FrozenSet, Mapping

from immutables import Map

from HABApp.openhab.items.base_item import OpenhabItem, MetaData
from HABApp.openhab.items.commands import UpDownCommand, PercentCommand
from ..definitions import UpDownValue, PercentValue


class RollershutterItem(OpenhabItem, UpDownCommand, PercentCommand):
    """RollershutterItem which accepts and converts the data types from OpenHAB

    :ivar str name:
    :ivar Union[int, float] value:

    :ivar Optional[str] label:
    :ivar FrozenSet[str] tags:
    :ivar FrozenSet[str] groups:
    :ivar Mapping[str, MetaData] metadata:
    """

    def __init__(self, name: str, initial_value=None, label: Optional[str] = None, tags: FrozenSet[str] = frozenset(),
                 groups: FrozenSet[str] = frozenset(), metadata: Mapping[str, MetaData] = Map()):
        if initial_value is not None:
            initial_value = float(initial_value)
        super().__init__(name, initial_value, label, tags, groups, metadata)

    def set_value(self, new_value) -> bool:

        if isinstance(new_value, UpDownValue):
            new_value = 0 if new_value.up else 100
        elif isinstance(new_value, PercentValue):
            new_value = new_value.value

        assert isinstance(new_value, (int, float)) or new_value is None, new_value
        return super().set_value(new_value)

    def is_up(self) -> bool:
        return self.value <= 0

    def is_down(self) -> bool:
        return self.value >= 100

    def __str__(self):
        return self.value



import re
from typing import List

from pydantic import BaseModel, Extra, Field

from HABApp.core.const import yml


class FileProperties(BaseModel):
    depends_on: List[str] = Field(alias='depends on', default_factory=list)
    reloads_on: List[str] = Field(alias='reloads on', default_factory=list)

    class Config:
        extra = Extra.forbid
        allow_population_by_field_name = True


RE_START = re.compile(r'^#(\s*)HABApp\s*:', re.IGNORECASE)


def get_properties(_str: str) -> FileProperties:

    cfg = []
    cut = 0

    # extract the property string
    for line in _str.splitlines():
        line = line.strip()
        if cut and not line:
            break

        if not line:
            continue

        # break on first non-empty line that is not a comment
        if line and not line.startswith('#'):
            break

        if not cut:
            # find out how much from the start we have to cut
            m = RE_START.search(line)
            if m:
                cut = len(m.group(1)) + 1
                cfg.append(line[cut:].lower())
        else:
            do_break = False
            for i, c in enumerate(line):
                if i > cut:
                    break

                if c not in ('#', ' ', '\t'):
                    do_break = True
                    break
            if do_break:
                break

            cfg.append(line[cut:])

    data = yml.load('\n'.join(cfg))
    if data is None:
        data = {}
    return FileProperties.parse_obj(data.get('habapp', {}))



# 1. Static stuff
from .__version__ import __version__

# 2. Setup used libraries and check installation
import HABApp.__setup_packages__

# 3. User configuration
import HABApp.config

# 4. Core features
import HABApp.core

# This holds only textual references to other objects so we can import this before everything else
import HABApp.rule_ctx


# Import the rest
import HABApp.mqtt
import HABApp.openhab
import HABApp.rule
import HABApp.runtime


import HABApp.util
from HABApp.rule import Rule
from HABApp.parameters import Parameter, DictParameter

from HABApp.config import CONFIG



from .handler import MidnightRotatingFileHandler

# isort: split

from .config import get_logging_dict, rotate_files, inject_log_buffer
from .default_logfile import get_default_logfile, create_default_logfile
from .queue_handler import HABAppQueueHandler



from typing import Any
from typing import Mapping

from immutables import Map
from pendulum import UTC
from pendulum import now as pd_now

from HABApp.core.items import BaseItem
from HABApp.openhab.events import ThingStatusInfoEvent, ThingUpdatedEvent, ThingConfigStatusInfoEvent
from HABApp.openhab.interface import set_thing_enabled


class Thing(BaseItem):
    """Base class for Things

    :ivar str status: Status of the thing (e.g. OFFLINE, ONLINE, ...)
    :ivar str status_detail: Additional detail for the status
    :ivar str label: Thing label
    :ivar Mapping[str, Any] configuration: Thing configuration
    :ivar Mapping[str, Any] properties: Thing properties
    """
    def __init__(self, name: str):
        super().__init__(name)

        self.status: str = ''
        self.status_detail: str = ''

        self.label: str = ''

        self.configuration: Mapping[str, Any] = Map()
        self.properties: Mapping[str, Any] = Map()

    def __update_timestamps(self, changed: bool):
        _now = pd_now(UTC)
        self._last_update.set(_now)
        if changed:
            self._last_change.set(_now)
        return None

    def process_event(self, event):

        if isinstance(event, ThingStatusInfoEvent):
            old = self.status
            self.status = new = event.status
            self.status_detail = event.detail

            self.__update_timestamps(old != new)
        elif isinstance(event, ThingUpdatedEvent):
            old_label         = self.label
            old_configuration = self.configuration
            old_properties    = self.properties

            self.label         = event.label
            self.configuration = Map(event.configuration)
            self.properties    = Map(event.properties)

            self.__update_timestamps(
                old_label != self.label or old_configuration != self.configuration or old_properties != self.properties
            )
        elif isinstance(event, ThingConfigStatusInfoEvent):
            pass

        return None

    def set_enabled(self, enable: bool = True):
        """Enable/disable the thing

        :param enable: True to enable, False to disable the thing
        :return:
        """
        return set_thing_enabled(self.name, enable)



import base64
import typing

from HABApp.core.events import ComplexEventValue


class OnOffValue(ComplexEventValue):
    ON = 'ON'
    OFF = 'OFF'

    def __init__(self, value):
        super().__init__(value)
        assert value == OnOffValue.ON or value == OnOffValue.OFF, f'{value} ({type(value)})'
        self.on = value == 'ON'

    def __str__(self):
        return self.value


class PercentValue(ComplexEventValue):
    def __init__(self, value: str):
        percent = float(value)
        assert 0 <= percent <= 100, f'{percent} ({type(percent)})'
        super().__init__(percent)

    def __str__(self):
        return f'{self.value}%'


class OpenClosedValue(ComplexEventValue):
    OPEN = 'OPEN'
    CLOSED = 'CLOSED'

    def __init__(self, value):
        super().__init__(value)
        assert value == OpenClosedValue.OPEN or value == OpenClosedValue.CLOSED, f'{value} ({type(value)})'
        self.open = value == OpenClosedValue.OPEN

    def __str__(self):
        return self.value


class UpDownValue(ComplexEventValue):
    UP = 'UP'
    DOWN = 'DOWN'

    def __init__(self, value):
        super().__init__(value)
        assert value == UpDownValue.UP or value == UpDownValue.DOWN, f'{value} ({type(value)})'
        self.up = value == UpDownValue.UP

    def __str__(self):
        return self.value


class HSBValue(ComplexEventValue):
    def __init__(self, value: str):
        super().__init__(tuple(float(k) for k in value.split(',')))

    def __str__(self):
        return f'{self.value[0]}°,{self.value[1]}%,{self.value[2]}%'


class QuantityValue(ComplexEventValue):

    @staticmethod
    def split_unit(value: str) -> typing.Tuple[str, str]:
        p = value.rfind(' ')
        # dimensionless has no unit
        if p < 0:
            return value, ''
        val = value[0:p]
        unit = value[p + 1:]
        return val, unit

    def __init__(self, value: str):
        value, unit = QuantityValue.split_unit(value)
        try:
            val: typing.Union[int, float] = int(value)
        except ValueError:
            val = float(value)
        super().__init__(val)
        self.unit = unit

    def __str__(self):
        return f'{self.value} {self.unit}'


class RawValue(ComplexEventValue):
    def __init__(self, value: str):
        # The data is in this format
        # data:image/png;base64,iVBORw0KGgo....

        # extract the contents from "data:"
        sep_type = value.find(';')
        self.type = value[5: sep_type]

        # this is our encoded payload
        sep_enc = value.find(',', sep_type)
        encoding = value[sep_type + 1: sep_enc]
        assert encoding == 'base64', f'"{encoding}"'

        # set the bytes as value
        super().__init__(base64.b64decode(value[sep_enc + 1:]))

    def __str__(self):
        return f'{self.type}'



from asyncio import create_task, Task
from typing import Callable, Awaitable, Any, Final, Optional


class SingleTask:
    def __init__(self, coro: Callable[[], Awaitable[Any]], name: Optional[str] = None):
        if name is None:
            name = f'{self.__class__.__name__}_{coro.__name__}'

        self.coro: Final = coro
        self.task: Optional[Task] = None
        self.name: Final = name

    def cancel(self):
        if self.task is not None:
            task = self.task
            self.task = None
            task.cancel()

    def start(self):
        self.cancel()
        self.task = create_task(self._task_wrap(), name=self.name)

    async def _task_wrap(self):
        # don't use try-finally because this also catches the asyncio.CancelledError
        try:
            await self.coro()
        except Exception:
            pass

        self.task = None



from datetime import timedelta
from typing import Union

TH_POSITIVE_TIME_DIFF = Union[int, float, timedelta]


def get_positive_time_diff(arg: TH_POSITIVE_TIME_DIFF, round_digits=None) -> Union[int, float]:
    if isinstance(arg, timedelta):
        diff = arg.total_seconds()
        if round_digits is not None:
            diff = round(diff, round_digits)
    elif isinstance(arg, float):
        diff = round(arg, round_digits) if round_digits is not None else arg
    elif isinstance(arg, int):
        diff = arg
    else:
        raise ValueError(f'Invalid type: {type(arg)}')

    if diff <= 0:
        raise ValueError(f'Time difference must be positive! ({arg})')
    return diff



import sys
from typing import Optional, Union, TYPE_CHECKING

from HABApp.core.errors import ContextNotSetError, ContextNotFoundError
from HABApp.core.internals.context import ContextProvidingObj, ContextBoundObj, HINT_CONTEXT_OBJ

if TYPE_CHECKING:
    import HABApp
    import types


def get_current_context(obj: Optional[ContextProvidingObj] = None) -> 'HABApp.rule_ctx.HABAppRuleContext':
    if obj is not None:
        return obj._habapp_ctx

    depth = 0
    while True:
        depth += 1
        try:
            frm = sys._getframe(depth)  # type: types.FrameType
        except ValueError:
            raise ContextNotFoundError() from None

        ctx_obj: Union[None, object, ContextProvidingObj] = frm.f_locals.get('self', None)
        if ctx_obj is None or not isinstance(ctx_obj, ContextProvidingObj):
            continue

        ctx = ctx_obj._habapp_ctx
        if ctx is None:
            raise ContextNotSetError()
        return ctx


class AutoContextBoundObj(ContextBoundObj):
    def __init__(self, parent_ctx: Optional['HINT_CONTEXT_OBJ'] = None, **kwargs):
        if parent_ctx is None:
            parent_ctx = get_current_context()
        super().__init__(parent_ctx=parent_ctx, **kwargs)



from .base_item import OpenhabItem
from .contact_item import ContactItem
from .dimmer_item import DimmerItem
from .rollershutter_item import RollershutterItem
from .switch_item import SwitchItem
from .color_item import ColorItem
from .number_item import NumberItem
from .datetime_item import DatetimeItem
from .string_item import StringItem, LocationItem, PlayerItem
from .call_item import CallItem
from .image_item import ImageItem
from .group_item import GroupItem
from .thing_item import Thing



import ruamel.yaml

yml = ruamel.yaml.YAML()
yml.default_flow_style = False
yml.default_style = False    # type: ignore
yml.width = 1000000          # type: ignore
yml.allow_unicode = True
yml.sort_base_mapping_type_on_output = False     # type: ignore



import sys
import time
from enum import Enum
from typing import Final


class _MissingType(Enum):
    MISSING = object()

    def __repr__(self):
        return '<Missing>'


MISSING: Final = _MissingType.MISSING
STARTUP: Final = time.monotonic()

# Python Versions for feature control
PYTHON_38: Final = sys.version_info >= (3, 8)
PYTHON_39: Final = sys.version_info >= (3, 9)
PYTHON_310: Final = sys.version_info >= (3, 10)
PYTHON_311: Final = sys.version_info >= (3, 11)



from typing import Dict, Type
from HABApp.core.const.json import load_json

from .events import OpenhabEvent, \
    ItemStateEvent, ItemStateChangedEvent, ItemCommandEvent, ItemAddedEvent, \
    ItemUpdatedEvent, ItemRemovedEvent, ItemStatePredictedEvent, GroupItemStateChangedEvent, \
    ChannelTriggeredEvent, ChannelDescriptionChangedEvent, \
    ThingAddedEvent, ThingRemovedEvent, ThingUpdatedEvent, \
    ThingStatusInfoChangedEvent, ThingStatusInfoEvent, ThingFirmwareStatusInfoEvent, \
    ThingConfigStatusInfoEvent


EVENT_LIST = [
    # item events
    ItemStateEvent, ItemStateChangedEvent, ItemCommandEvent, ItemAddedEvent,
    ItemUpdatedEvent, ItemRemovedEvent, ItemStatePredictedEvent, GroupItemStateChangedEvent,

    # channel events
    ChannelTriggeredEvent, ChannelDescriptionChangedEvent,

    # thing events
    ThingAddedEvent, ThingRemovedEvent, ThingUpdatedEvent,
    ThingStatusInfoEvent, ThingStatusInfoChangedEvent,
    ThingFirmwareStatusInfoEvent,
    ThingConfigStatusInfoEvent,
]

_events: Dict[str, Type[OpenhabEvent]] = {k.__name__: k for k in EVENT_LIST}
_events['FirmwareStatusInfoEvent'] = ThingFirmwareStatusInfoEvent    # Naming from openHAB is inconsistent here
_events['ConfigStatusInfoEvent'] = ThingConfigStatusInfoEvent        # Naming from openHAB is inconsistent here


def get_event(_in_dict: dict) -> OpenhabEvent:
    event_type: str = _in_dict['type']
    topic: str = _in_dict['topic']

    # Workaround for None values in the payload str
    p_str: str = _in_dict['payload']
    if '"NONE"' in p_str:
        p_str = p_str.replace('"NONE"', 'null')
    payload = load_json(p_str)

    # Find event from implemented events
    try:
        return _events[event_type].from_dict(topic, payload)
    except KeyError:
        raise ValueError(f'Unknown Event: {event_type:s} for {_in_dict}')



from __future__ import annotations

import logging
import typing
from pathlib import Path
from typing import Callable, Awaitable, Any

from HABApp.core.files.errors import CircularReferenceError, DependencyDoesNotExistError, AlreadyHandledFileError
from HABApp.core.files.file.properties import FileProperties
from HABApp.core.files.manager.files import FILES, file_state_changed
from HABApp.core.wrapper import process_exception
from . import FileState

log = logging.getLogger('HABApp.files')


class HABAppFile:
    LOGGER: logging.Logger
    LOAD_FUNC: Callable[[str, Path], Awaitable[Any]]
    UNLOAD_FUNC: Callable[[str, Path], Awaitable[Any]]

    def __init__(self, name: str, path: Path, properties: FileProperties):
        self.name: str = name
        self.path: Path = path

        self.state: FileState = FileState.PENDING
        self.properties: FileProperties = properties
        log.debug(f'{self.name} added')

    def __repr__(self):
        return f'<{self.__class__.__name__} {self.name} state: {self.state}>'

    def set_state(self, new_state: FileState):
        if self.state is new_state:
            return None

        self.state = new_state
        log.debug(f'{self.name} changed to {self.state}')
        file_state_changed(self)

    def _check_circ_refs(self, stack, prop: str):
        c: typing.List[str] = getattr(self.properties, prop)
        for f in c:
            _stack = stack + (f, )
            if f in stack:
                raise CircularReferenceError(_stack)

            next_file = FILES.get(f)
            if next_file is not None:
                next_file._check_circ_refs(_stack, prop)

    def _check_properties(self):
        # check dependencies
        mis = set(filter(lambda x: x not in FILES, self.properties.depends_on))
        if mis:
            one = len(mis) == 1
            msg = f'File {self.path} depends on file{"" if one else "s"} that ' \
                  f'do{"es" if one else ""}n\'t exist: {", ".join(sorted(mis))}'

            raise DependencyDoesNotExistError(msg)

        # check reload
        mis = set(filter(lambda x: x not in FILES, self.properties.reloads_on))
        if mis:
            one = len(mis) == 1
            log.warning(f'File {self.path} reloads on file{"" if one else "s"} that '
                        f'do{"es" if one else ""}n\'t exist: {", ".join(sorted(mis))}')

    def check_properties(self, log_msg: bool = False):
        if self.state is not FileState.PENDING and self.state is not FileState.DEPENDENCIES_ERROR:
            return None

        try:
            self._check_properties()
        except DependencyDoesNotExistError as e:
            if log_msg:
                log.error(e.msg)
            return self.set_state(FileState.DEPENDENCIES_ERROR)

        try:
            # check for circular references
            self._check_circ_refs((self.name, ), 'depends_on')
            self._check_circ_refs((self.name, ), 'reloads_on')
        except CircularReferenceError as e:
            log.error(f'Circular reference: {" -> ".join(e.stack)}')
            return self.set_state(FileState.DEPENDENCIES_ERROR)

        # Check if we can already load it
        self.set_state(FileState.DEPENDENCIES_OK if not self.properties.depends_on else FileState.DEPENDENCIES_MISSING)

    def check_dependencies(self):
        if self.state is not FileState.DEPENDENCIES_MISSING:
            return None

        for name in self.properties.depends_on:
            f = FILES.get(name, None)
            if f is None:
                return None
            if f.state is not FileState.LOADED:
                return None

        self.set_state(FileState.DEPENDENCIES_OK)
        return None

    async def load(self):
        assert self.state is FileState.DEPENDENCIES_OK, self.state

        try:
            await self.__class__.LOAD_FUNC(self.name, self.path)
        except Exception as e:
            if not isinstance(e, AlreadyHandledFileError):
                process_exception(self.__class__.LOAD_FUNC, e, logger=self.LOGGER)
            self.set_state(FileState.FAILED)
            return None

        self.set_state(FileState.LOADED)
        return None

    async def unload(self):
        try:
            await self.__class__.UNLOAD_FUNC(self.name, self.path)
        except Exception as e:
            if not isinstance(e, AlreadyHandledFileError):
                process_exception(self.__class__.UNLOAD_FUNC, e, logger=self.LOGGER)
            self.set_state(FileState.FAILED)
            return None

        self.set_state(FileState.REMOVED)
        return None

    def file_changed(self, file: HABAppFile):
        name = file.name
        if name in self.properties.reloads_on:
            self.set_state(FileState.PENDING)



from typing import FrozenSet, Mapping, Optional

from immutables import Map

from HABApp.openhab.items.base_item import OpenhabItem, MetaData


class CallItem(OpenhabItem):
    """CallItem which accepts and converts the data types from OpenHAB

    :ivar str name:
    :ivar Tuple[str, ...] value:

    :ivar Optional[str] label:
    :ivar FrozenSet[str] tags:
    :ivar FrozenSet[str] groups:
    :ivar Mapping[str, MetaData] metadata:
    """

    @classmethod
    def from_oh(cls, name: str, value=None, label: Optional[str] = None, tags: FrozenSet[str] = frozenset(),
                groups: FrozenSet[str] = frozenset(), metadata: Mapping[str, MetaData] = Map()):
        if value is not None:
            value = tuple(value.split(','))
        return cls(name, value, label=label, tags=tags, groups=groups, metadata=metadata)

    def set_value(self, new_value) -> bool:
        if isinstance(new_value, str):
            new_value = tuple(new_value.split(','))
        return super().set_value(new_value)



from pydantic import BaseModel, Extra, validator


class RestBase(BaseModel):

    # default configuration for RestAPI models
    class Config:
        extra = Extra.forbid


def none_is_empty_str(v) -> str:
    return None if v == 'NONE' else v


def make_none_empty_str(*name: str):
    return validator(*name, allow_reuse=True)(none_is_empty_str)



from .proxy import uses_get_item, uses_item_registry, uses_post_event, uses_event_bus, setup_internals
from .context import ContextProvidingObj, Context, ContextBoundObj, get_current_context, HINT_CONTEXT_OBJ, AutoContextBoundObj

# isort: split

from .event_filter import EventFilterBase, HINT_EVENT_FILTER_OBJ
from .event_bus import EventBus, HINT_EVENT_BUS
from .item_registry import HINT_ITEM_REGISTRY, ItemRegistry, ItemRegistryItem

# isort: split

from .event_bus_listener import HINT_EVENT_BUS_LISTENER, EventBusListener, ContextBoundEventBusListener
from .event_filter import EventFilterBase, HINT_EVENT_FILTER_OBJ
from .wrapped_function import TYPE_WRAPPED_FUNC_OBJ, wrap_func



from asyncio import create_task
from typing import Union

import HABApp
import HABApp.core
import HABApp.openhab.events
from HABApp.core.errors import ItemNotFoundException
from HABApp.core.events import ValueUpdateEvent, ValueChangeEvent
from HABApp.core.internals import uses_post_event, uses_get_item
from HABApp.core.logger import log_warning
from HABApp.core.wrapper import process_exception
from HABApp.openhab.connection_handler import http_connection
from HABApp.openhab.events import GroupItemStateChangedEvent, ItemAddedEvent, ItemRemovedEvent, ItemUpdatedEvent, \
    ThingStatusInfoEvent, ThingAddedEvent, ThingRemovedEvent, ThingUpdatedEvent, ThingConfigStatusInfoEvent
from HABApp.openhab.item_to_reg import add_to_registry, remove_from_registry, remove_thing_from_registry, \
    add_thing_to_registry
from HABApp.openhab.map_events import get_event
from HABApp.openhab.map_items import map_item
from HABApp.openhab.definitions.topics import TOPIC_THINGS, TOPIC_ITEMS

log = http_connection.log


post_event = uses_post_event()
get_item = uses_get_item()


def on_sse_event(event_dict: dict):
    try:
        # Lookup corresponding OpenHAB event
        event = get_event(event_dict)

        # Update item in registry BEFORE posting to the event bus
        # so the items have the correct state when we process the event in a rule
        try:
            if isinstance(event, ValueUpdateEvent):
                __item = get_item(event.name)  # type: HABApp.core.items.base_valueitem.BaseValueItem
                __item.set_value(event.value)
                post_event(event.name, event)
                return None

            if isinstance(event, ValueChangeEvent):
                post_event(event.name, event)
                return None

            if isinstance(event, (ThingStatusInfoEvent, ThingUpdatedEvent, ThingConfigStatusInfoEvent)):
                __thing = get_item(event.name)   # type: HABApp.openhab.items.Thing
                __thing.process_event(event)
                post_event(event.name, event)
                return None

            # Workaround because there is no GroupItemStateEvent
            if isinstance(event, GroupItemStateChangedEvent):
                __item = get_item(event.name)  # type: HABApp.openhab.items.GroupItem
                __item.set_value(event.value)
                post_event(event.name, event)
                return None
        except ItemNotFoundException:
            log_warning(log, f'Received {event.__class__.__name__} for {event.name} but item does not exist!')

            # Post the event anyway
            post_event(event.name, event)
            return None

        # Events that add items to the item registry
        # These events require that we query openHAB because of the metadata, so we have to do it in a task
        if isinstance(event, (ItemAddedEvent, ItemUpdatedEvent)):
            create_task(item_event(event))
            return None

        # Events that remove items from the item registry
        if isinstance(event, ItemRemovedEvent):
            remove_from_registry(event.name)
            post_event(TOPIC_ITEMS, event)
            return None

        # Events that add things to the item registry
        if isinstance(event, ThingAddedEvent):
            create_task(thing_event(event))
            return None

        # Events that remove things from the item registry
        if isinstance(event, ThingRemovedEvent):
            remove_thing_from_registry(event.name)
            post_event(TOPIC_THINGS, event)
            return None

        # Unknown Event -> just forward it to the event bus
        post_event(event.name, event)
    except Exception as e:
        process_exception(func=on_sse_event, e=e)
        return None


async def item_event(event: Union[ItemAddedEvent, ItemUpdatedEvent]):
    name = event.name

    # Since metadata is not part of the event we have to request it
    cfg = await HABApp.openhab.interface_async.async_get_item(name, metadata='.+')

    new_item = map_item(name, event.type, None, event.label, event.tags, event.groups, metadata=cfg.get('metadata'))
    if new_item is None:
        return None

    add_to_registry(new_item)
    # Send Event to Event Bus
    post_event(TOPIC_ITEMS, event)
    return None


async def thing_event(event: ThingAddedEvent):
    # Since the thing status is not part of the event we have to request ist
    add_thing_to_registry(await HABApp.openhab.interface_async.async_get_thing(event.name))

    # Send Event to Event Bus
    post_event(TOPIC_THINGS, event)
    return None



from typing import Optional, TypeVar

from HABApp.core.internals.event_bus import EventBusBaseListener
from HABApp.core.internals.wrapped_function import TYPE_WRAPPED_FUNC_OBJ, WrappedFunctionBase
from HABApp.core.internals import uses_event_bus, HINT_CONTEXT_OBJ
from HABApp.core.internals import HINT_EVENT_FILTER_OBJ, AutoContextBoundObj


event_bus = uses_event_bus()


class EventBusListener(EventBusBaseListener):
    def __init__(self, topic: str, callback: TYPE_WRAPPED_FUNC_OBJ, event_filter: HINT_EVENT_FILTER_OBJ, **kwargs):
        super().__init__(topic, **kwargs)

        assert isinstance(callback, WrappedFunctionBase)
        self.func: TYPE_WRAPPED_FUNC_OBJ = callback
        self.filter: HINT_EVENT_FILTER_OBJ = event_filter

    def notify_listeners(self, event):
        if self.filter.trigger(event):
            self.func.run(event)

    def describe(self) -> str:
        return f'"{self.topic}" (filter={self.filter.describe()})'

    def cancel(self):
        """Stop listening on the event bus"""
        event_bus.remove_listener(self)


HINT_EVENT_BUS_LISTENER = TypeVar('HINT_EVENT_BUS_LISTENER', bound=EventBusListener)


class ContextBoundEventBusListener(EventBusListener, AutoContextBoundObj):
    def __init__(self, topic: str, callback: TYPE_WRAPPED_FUNC_OBJ, event_filter: HINT_EVENT_FILTER_OBJ,
                 parent_ctx: Optional[HINT_CONTEXT_OBJ] = None):
        super().__init__(topic=topic, callback=callback, event_filter=event_filter, parent_ctx=parent_ctx)

        assert isinstance(callback, WrappedFunctionBase)
        self.func: TYPE_WRAPPED_FUNC_OBJ = callback
        self.filter: HINT_EVENT_FILTER_OBJ = event_filter

    def notify_listeners(self, event):
        if self.filter.trigger(event):
            self.func.run(event)

    def describe(self) -> str:
        return f'"{self.topic}" (filter={self.filter.describe()})'

    def _ctx_unlink(self):
        event_bus.remove_listener(self)
        return super()._ctx_unlink()

    def cancel(self):
        """Stop listening on the event bus"""
        self._ctx_unlink()



class EventBusBaseListener:
    def __init__(self, topic: str, **kwargs):
        super().__init__(**kwargs)
        assert isinstance(topic, str)
        self.topic: str = topic

    def notify_listeners(self, event):
        raise NotImplementedError()

    def describe(self) -> str:
        raise NotImplementedError()



from asyncio import Future as _Future
from asyncio import run_coroutine_threadsafe as _run_coroutine_threadsafe
from contextvars import ContextVar as _ContextVar
from typing import Any as _Any
from typing import Callable as _Callable
from typing import Coroutine as _Coroutine
from typing import Optional as _Optional
from typing import TypeVar as _TypeVar

from HABApp.core.const import loop

async_context = _ContextVar('async_ctx')


class AsyncContextError(Exception):
    def __init__(self, func: _Callable) -> None:
        super().__init__()
        self.func: _Callable = func

    def __str__(self):
        return f'Function "{self.func.__name__}" may not be called from an async context!'


def create_task(coro: _Coroutine, name: _Optional[str] = None) -> _Future:
    if async_context.get(None) is None:
        return _run_coroutine_threadsafe(coro, loop)
    else:
        return loop.create_task(coro, name=name)


_CORO_RET = _TypeVar('_CORO_RET')


def run_coro_from_thread(coro: _Coroutine[_Any, _Any, _CORO_RET], calling: _Callable) -> _CORO_RET:
    # This function call is blocking, so it can't be called in the async context
    if async_context.get(None) is not None:
        raise AsyncContextError(calling)

    fut = _run_coroutine_threadsafe(coro, loop)
    return fut.result()



import logging
import typing

from HABApp.core.internals import uses_post_event
from HABApp.core.lib import format_exception
from HABApp.core.const.topics import TOPIC_ERRORS as _T_ERRORS
from HABApp.core.const.topics import TOPIC_INFOS as _T_INFOS
from HABApp.core.const.topics import TOPIC_WARNINGS as _T_WARNINGS


post_event = uses_post_event()


def log_error(logger: logging.Logger, text: str):
    if '\n' in text:
        for line in text.splitlines():
            logger.error(line)
    else:
        logger.error(text)
    post_event(
        _T_ERRORS, text
    )


def log_warning(logger: logging.Logger, text: str):
    if '\n' in text:
        for line in text.splitlines():
            logger.warning(line)
    else:
        logger.warning(text)

    post_event(
        _T_WARNINGS, text
    )


def log_info(logger: logging.Logger, text: str):
    if '\n' in text:
        for line in text.splitlines():
            logger.info(line)
    else:
        logger.info(text)

    post_event(
        _T_INFOS, text
    )


class HABAppLogger:
    _LEVEL: int
    _TOPIC: str

    def __init__(self, log: logging.Logger):
        self.lines: typing.List[str] = []
        self.logger = log

    def add(self, text: str, *args, **kwargs):
        self.lines.append(text.format(*args, **kwargs))
        return self

    def add_exception(self, e: Exception, add_traceback: bool = False):
        if not add_traceback:
            for line in str(e).splitlines():
                self.lines.append(line)
        else:
            self.lines.extend(format_exception(e))
        return self

    def dump(self) -> bool:
        if not self.lines:
            return False

        if self.logger.isEnabledFor(self._LEVEL):
            for line in self.lines:
                self.logger._log(self._LEVEL, line, [])

        post_event(
            self._TOPIC, '\n'.join(self.lines)
        )
        self.lines.clear()
        return True

    def __bool__(self):
        return bool(self.lines)


class HABAppError(HABAppLogger):
    _LEVEL = logging.ERROR
    _TOPIC = _T_ERRORS


class HABAppWarning(HABAppLogger):
    _LEVEL = logging.WARNING
    _TOPIC = _T_WARNINGS


class HABAppInfo(HABAppLogger):
    _LEVEL = logging.INFO
    _TOPIC = _T_INFOS



from pathlib import Path
from typing import Optional


def get_log_folder(default: Optional[Path] = None) -> Optional[Path]:
    # As a default we log into the openHAB folder
    choices = ('/var/log/openhab', '/opt/openhab/userdata/logs')
    for choice in choices:
        path = Path(choice)
        if path.is_dir():
            return path

    return default


def is_openhabian():
    return Path('/opt/openhabian').is_dir()



from HABApp.core import const
from HABApp.core import lib
from HABApp.core import errors

# isort: split

from HABApp.core import asyncio

# isort: split
from HABApp.core import internals

from HABApp.core import wrapper
from HABApp.core import logger

# isort: split

import HABApp.core.events
import HABApp.core.files
import HABApp.core.items

# isort: split

Items: 'HABApp.core.internals.ItemRegistry' = internals.proxy.ConstProxyObj('ItemRegistry')
EventBus: 'HABApp.core.internals.EventBus' = internals.proxy.ConstProxyObj('EventBus')



from datetime import date, datetime
from logging.handlers import RotatingFileHandler


class MidnightRotatingFileHandler(RotatingFileHandler):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.last_check: date = datetime.now().date()

    def shouldRollover(self, record):
        date = datetime.now().date()
        if date == self.last_check:
            return 0
        self.last_check = date
        return super().shouldRollover(record)



from collections import OrderedDict
from typing import Optional, Dict, List, Union, Tuple, Any


class Column:
    wrap: int = 80

    def __init__(self, name: str, align: Optional[str] = None, alias: Optional[str] = None, wrap: Optional[int] = None):
        self.name: str = name
        self.alias: Optional[str] = alias

        self.align: Optional[str] = align
        if wrap is not None:
            self.wrap = wrap

        self.width: int = len(name) if alias is None else len(alias)
        self.entries: List[Tuple[Any, ...]] = []

    def get_lines(self, pos: int) -> int:
        return len(self.entries[pos])

    def format_entry(self, pos: int, lines: int) -> List[str]:
        ret = []

        objs = self.entries[pos]
        size = len(objs)
        for i in range(lines):
            if i >= size:
                ret.append(self.width * ' ')
                continue
            val = objs[i]
            if isinstance(val, bool):
                val = str(val)
            f = f'{{:{""if self.align is None else self.align}{self.width:d}}}'
            ret.append(f.format(val))
        return ret

    def add(self, val):
        _res = []
        if isinstance(val, (list, set, tuple)):
            _len = 0
            _str = ''
            for obj in val:
                if _len >= self.wrap:
                    _len = 0
                    _res.append(_str)
                    _str = ''
                _str = f'{_str}, {obj}' if _str else f'{obj}'
                _len = len(_str)
            _res.append(_str)
        else:
            _res.append(val)

        for k in _res:
            self.width = max(self.width, len(str(k)))
        self.entries.append(tuple(_res))


class Table:
    def __init__(self, heading: str = ''):
        self.columns: Dict[str, Column] = OrderedDict()
        self.heading: str = heading

    def add_column(self, name: str, align: Optional[str] = None, alias: Optional[str] = None,
                   wrap: Optional[int] = None) -> Column:
        self.columns[name] = c = Column(name, align, alias, wrap)
        return c

    def add_dict(self, _in: dict):
        for k, col in self.columns.items():
            col.add(_in[k])

    def get_lines(self, sort_columns: List[Union[str, Column]] = None) -> List[str]:
        # check if all tables have the same length
        vals = list(self.columns.values())
        len1 = len(vals[0].entries)
        assert all(map(lambda x: len(x.entries) == len1, vals)), {k: len(v.entries) for k, v in self.columns.items()}

        # We don't show the empty table
        if len1 <= 0:
            return []

        # Sort entries
        if sort_columns is None:
            lines_dict = {i: i for i in range(len1)}
        else:
            # Convert column names to columns
            for i, e in enumerate(sort_columns):
                if isinstance(e, str):
                    sort_columns[i] = self.columns[e]

            lines_dict = {}
            for i in range(len1):
                lines_dict[tuple(c.entries[i] for c in sort_columns) + (i,)] = i

        line_sep = '+-' + '-+-'.join(map(lambda x: '-' * x.width, self.columns.values())) + '-+'

        ret = []
        if self.heading:
            # this is a cheap hacky solution to quickly get the width
            h_width = len(line_sep) - 2
            ret.append('+' + '-' * h_width + '+')
            ret.append(f'|{self.heading:^{h_width}s}|')

        ret.append(line_sep)

        # Heading
        l1 = '|'
        for v in self.columns.values():
            l1 += f' {v.name if v.alias is None else v.alias:^{v.width}s} |'
        ret.append(l1)
        ret.append(line_sep)

        for t, i in sorted(lines_dict.items()):
            lines = max(map(lambda x: x.get_lines(i), self.columns.values()))

            grid = tuple(map(lambda x: x.format_entry(i, lines), self.columns.values()))  # type: Tuple[List[str], ...]
            for col_i in range(lines):
                cols = [obj[col_i] for obj in grid]
                ret.append('| ' + ' | '.join(cols) + ' |')

        ret.append(line_sep)
        return ret



import logging
from typing import Optional, Callable

import HABApp
from HABApp.core.const.topics import ALL_TOPICS
from HABApp.core.internals import Context, uses_item_registry, HINT_EVENT_BUS_LISTENER
from HABApp.core.internals import uses_event_bus
from HABApp.core.internals.event_bus import EventBusBaseListener

event_bus = uses_event_bus()
item_registry = uses_item_registry()

log = logging.getLogger('HABApp.Rule')


class HABAppRuleContext(Context):
    def __init__(self, rule: 'HABApp.rule.Rule'):
        super().__init__()
        self.rule: Optional['HABApp.rule.Rule'] = rule

    def get_callback_name(self, callback: Callable) -> Optional[str]:
        return f'{self.rule.rule_name}.{callback.__name__}' if self.rule.rule_name else None

    def add_event_listener(self, listener: HINT_EVENT_BUS_LISTENER) -> HINT_EVENT_BUS_LISTENER:
        event_bus.add_listener(listener)
        return listener

    def remove_event_listener(self, listener: HINT_EVENT_BUS_LISTENER) -> HINT_EVENT_BUS_LISTENER:
        event_bus.remove_listener(listener)
        return listener

    def unload_rule(self):
        with HABApp.core.wrapper.ExceptionToHABApp(log):
            rule = self.rule

            # Unload the scheduler
            rule.run._scheduler.cancel_all()
            rule.run._habapp_ctx = None

            # cancel things and set obj to None
            while self.objs:
                with HABApp.core.wrapper.ExceptionToHABApp(log):
                    to_cancel = next(iter(self.objs))
                    to_cancel.cancel()
            self.objs = None    # Set to None so we crash if we want to schedule new stuff

            # clean references
            self.rule = None
            rule._habapp_rule_ctx = None

            # user implementation
            rule.on_rule_removed()

    def check_rule(self):
        with HABApp.core.wrapper.ExceptionToHABApp(log):
            # We need items if we want to run the test
            if item_registry.get_items():

                # Check if we have a valid item for all listeners
                for listener in self.objs:
                    if not isinstance(listener, EventBusBaseListener):
                        continue

                    # Internal topics - don't warn there
                    if listener.topic in ALL_TOPICS:
                        continue

                    # check if specific item exists
                    if not item_registry.item_exists(listener.topic):
                        log.warning(f'Item "{listener.topic}" does not exist (yet)! '
                                    f'self.listen_event in "{self.rule.rule_name}" may not work as intended.')

            # enable the scheduler
            self.rule.run._scheduler.resume()

            # user implementation
            self.rule.on_rule_loaded()



from typing import TypeVar, Any


class EventFilterBase:
    def trigger(self, event: Any) -> bool:
        raise NotImplementedError()

    def describe(self) -> str:
        raise NotImplementedError()

    def __repr__(self):
        return f'<{self.describe()} at 0x{id(self):X}>'


# Hints for functions that use an item class as an input parameter
HINT_EVENT_FILTER_OBJ = TypeVar('HINT_EVENT_FILTER_OBJ', bound=EventFilterBase)



import re
from typing import Optional, Dict, Tuple, Pattern, Union

from .filters import THING_ALIAS, CHANNEL_ALIAS


SEPARATOR = ','
RE_PLACEHOLDERS = re.compile(r'{([^}]+)}')
RE_ACCESSOR = re.compile(f'(?P<accessor>[^{SEPARATOR}]+)(?:{SEPARATOR}(?P<operator>.+))?')


class StrBuilder:
    def __init__(self, value: str):
        self.template: str = value.strip()
        self.regex: Dict[str, Tuple[str, Optional[Pattern], Optional[str]]] = {}

        for p in RE_PLACEHOLDERS.findall(value):
            m = RE_ACCESSOR.fullmatch(p)
            accessor = m.group('accessor')
            operator = m.group('operator')

            regex, replace = None, None

            if operator is not None:
                p_comma = operator.rfind(SEPARATOR)
                if p_comma == -1:
                    search = operator
                else:
                    search = operator[:p_comma]
                    replace = operator[p_comma + 1:].strip()

                try:
                    search = search.strip()
                    regex = re.compile(search, re.IGNORECASE)
                    if replace is None and not regex.groups:
                        raise ValueError(f'Pattern {search} does not have a group!')
                except re.error as e:
                    raise ValueError(f'Could not compile regex "{search}": {e}') from None

            assert accessor in THING_ALIAS or accessor in CHANNEL_ALIAS, accessor
            self.regex[f'{{{p}}}'] = accessor.strip(), regex, replace

    def get_str(self, context: Dict[str, str]) -> str:
        out = self.template
        for search, p in self.regex.items():
            accessor, regex, replace = p

            if regex is None:
                out = out.replace(search, context[accessor])
                continue

            if replace is not None:
                val = regex.sub(replace, context[accessor])
            else:
                m = regex.search(context[accessor])
                if m is None:
                    raise ValueError(f'Regex "{regex.pattern}" did not match "{context[accessor]}"')
                val = m.group(1)
            out = out.replace(search, val.strip())

        return out

    def __repr__(self):
        return f'<{self.__class__.__name__} "{self.template}">'


def apply_builder(_in: Union[dict, list], context: dict):
    if isinstance(_in, StrBuilder):
        return _in.get_str(context)

    if isinstance(_in, list):
        ret = []
        for k in _in:
            if isinstance(k, StrBuilder):
                k = k.get_str(context)
            elif isinstance(k, (list, dict)):
                k = apply_builder(k, context)
            ret.append(k)
    else:
        ret = {}
        for k, v in _in.items():
            if isinstance(k, StrBuilder):
                k = k.get_str(context)

            if isinstance(v, StrBuilder):
                v = v.get_str(context)
            elif isinstance(v, (list, dict)):
                v = apply_builder(v, context)
            ret[k] = v
    return ret



import asyncio
import time


MAX_WAIT: int = 180


class WaitBetweenConnects:
    def __init__(self):
        self.last_try: float = 0
        self.wait_time = 0

    async def wait(self):

        # wenn wir lang connected sind oder beim ersten mal versuchen wir den reconnect gleich
        if time.time() - self.last_try > MAX_WAIT:
            wait = 0
        else:
            wait = self.wait_time
            wait = wait * 2 if wait <= 16 else wait + 8
            wait = max(wait, 1)
            wait = min(wait, MAX_WAIT)

        self.wait_time = wait
        await asyncio.sleep(self.wait_time)

        self.last_try = time.time()



import itertools
import logging
import logging.handlers
import traceback
import typing
import signal
from asyncio import iscoroutinefunction, sleep
from dataclasses import dataclass
from types import FunctionType, MethodType
from typing import Callable, Coroutine, Union

from HABApp.core.asyncio import async_context, create_task
from HABApp.core.const import loop


@dataclass(frozen=True)
class ShutdownInfo:
    func: Union[Callable[[], typing.Any], Coroutine]
    msg: str
    last: bool


_FUNCS: typing.List[ShutdownInfo] = []

requested: bool = False


def register_func(func, last=False, msg: str = ''):
    assert isinstance(func, (FunctionType, MethodType)) or iscoroutinefunction(func), print(type(func))
    assert last is True or last is False, last
    assert isinstance(msg, str)

    _FUNCS.append(ShutdownInfo(func, f'{func.__module__}.{func.__name__}' if not msg else msg, last))


def register_signal_handler():
    def shutdown_handler(sig, frame):
        print('Shutting down ...')
        request_shutdown()

    # register shutdown helper
    signal.signal(signal.SIGINT, shutdown_handler)
    signal.signal(signal.SIGTERM, shutdown_handler)


def request_shutdown():
    create_task(_shutdown())


async def _shutdown():
    global requested
    "Request execution of all functions"

    async_context.set('Shutdown')

    log = logging.getLogger('HABApp.Shutdown')
    log.debug('Requested shutdown')

    requested = True

    for obj in itertools.chain(filter(lambda x: not x.last, _FUNCS),
                               filter(lambda x: x.last, _FUNCS),
                               # shutdown of the event loop has to be the last thing that is done
                               # since stopping of the loop exits the program
                               [ShutdownInfo(loop.stop, 'Stopping asyncio loop', last=True)]):
        try:
            log.debug(f'{obj.msg}')
            if iscoroutinefunction(obj.func):
                await obj.func()
            else:
                obj.func()
            log.debug('-> done!')
            await sleep(0.02)
        except Exception as ex:
            log.error(ex)
            tb = traceback.format_exc().splitlines()
            for line in tb:
                log.error(line)

    log.debug('Shutdown complete')
    return None



import logging
import threading
from pathlib import Path

import HABApp
from HABApp.core.files.file import HABAppFile
from HABApp.core.files.folders import add_folder as add_habapp_folder
from .parameters import get_parameter_file, remove_parameter_file, set_parameter_file

log = logging.getLogger('HABApp.RuleParameters')

LOCK = threading.Lock()
PARAM_PREFIX = 'params/'


async def load_file(name: str, path: Path):
    with LOCK:  # serialize to get proper error messages
        with path.open(mode='r', encoding='utf-8') as file:
            data = HABApp.core.const.yml.load(file)
        if data is None:
            data = {}
        set_parameter_file(path.stem, data)

    log.debug(f'Loaded params from {name}!')


async def unload_file(name: str, path: Path):
    with LOCK:  # serialize to get proper error messages
        remove_parameter_file(path.stem)

    log.debug(f'Removed params from {path.name}!')


def save_file(file: str):
    assert isinstance(file, str), type(file)
    path = HABApp.CONFIG.directories.param
    if path is None:
        raise ValueError('Parameter files are disabled! Configure a folder to use them!')

    filename = path / (file + '.yml')

    with LOCK:  # serialize to get proper error messages
        log.info(f'Updated {filename}')
        with filename.open('w', encoding='utf-8') as outfile:
            HABApp.core.const.yml.dump(get_parameter_file(file), outfile)


class HABAppParameterFile(HABAppFile):
    LOGGER = log
    LOAD_FUNC = load_file
    UNLOAD_FUNC = unload_file


async def setup_param_files() -> bool:
    path = HABApp.CONFIG.directories.param
    if path is None:
        return False

    folder = add_habapp_folder(PARAM_PREFIX, path, 100)
    folder.add_file_type(HABAppParameterFile)
    watcher = folder.add_watch('.yml')
    await watcher.trigger_all()


def reload_param_file(name: str):
    name = f'{PARAM_PREFIX}{name}.yml'
    path = HABApp.core.files.folders.get_path(name)
    HABApp.core.asyncio.create_task(HABApp.core.files.manager.process_file(name, path))



from asyncio import run_coroutine_threadsafe, sleep
from pathlib import Path
from time import time
from typing import Any, List, Set, Awaitable, Callable

import HABApp
from HABApp.core.wrapper import ignore_exception
from .base_watcher import EventFilterBase
from .base_watcher import FileSystemEventHandler

DEBOUNCE_TIME: float = 0.6


class AggregatingAsyncEventHandler(FileSystemEventHandler):
    def __init__(self, folder: Path, func: Callable[[List[Path]], Awaitable[Any]], filter: EventFilterBase,
                 watch_subfolders: bool = False):
        super().__init__(folder, filter, watch_subfolders=watch_subfolders)

        self.func = func

        self._files: Set[Path] = set()
        self.last_event: float = 0

    @ignore_exception
    def file_changed(self, dst: str):
        # Map from thread to async
        run_coroutine_threadsafe(self._event_waiter(Path(dst)), loop=HABApp.core.const.loop)

    @ignore_exception
    async def _event_waiter(self, dst: Path):
        self.last_event = ts = time()
        self._files.add(dst)

        # debounce time
        await sleep(DEBOUNCE_TIME)

        # check if a new event came
        if self.last_event > ts:
            return None

        # Copy Path so we're done here
        files = list(self._files)
        self._files.clear()

        # process
        await self.func(HABApp.core.lib.sort_files(files))

    async def trigger_all(self):
        files = HABApp.core.lib.list_files(self.folder, self.filter, self.watch_subfolders)
        await self.func(files)



from typing import Any, Callable

try:
    import ujson
    load_json: Callable[[str], Any] = ujson.loads
    dump_json: Callable[[str], Any] = ujson.dumps
except ImportError:
    import json
    load_json: Callable[[str], Any] = json.loads
    dump_json: Callable[[str], Any] = json.dumps



from typing import Any

from HABApp.core.const import MISSING
from HABApp.core.events.filter.event import TypeBoundEventFilter
from . import ItemStateChangedEvent, ItemStateEvent, ItemCommandEvent


class ItemStateEventFilter(TypeBoundEventFilter):
    def __init__(self, value: Any = MISSING):
        super().__init__(ItemStateEvent, value=value)


class ItemStateChangedEventFilter(TypeBoundEventFilter):
    def __init__(self, value: Any = MISSING, old_value: Any = MISSING):
        super().__init__(ItemStateChangedEvent, value=value, old_value=old_value)


class ItemCommandEventFilter(TypeBoundEventFilter):
    def __init__(self, value: Any = MISSING):
        super().__init__(ItemCommandEvent, value=value)



from typing import Dict, TYPE_CHECKING

if TYPE_CHECKING:
    import HABApp

FILES: Dict[str, 'HABApp.core.files.file.HABAppFile'] = {}


def file_state_changed(file: 'HABApp.core.files.file.HABAppFile'):
    for f in FILES.values():
        if f is not file:
            f.file_changed(file)



from typing import Any, Tuple

from HABApp.core.internals import EventFilterBase, HINT_EVENT_FILTER_OBJ


class EventFilterBaseGroup(EventFilterBase):
    def __init__(self, *args: HINT_EVENT_FILTER_OBJ):
        self.filters: Tuple[HINT_EVENT_FILTER_OBJ, ...] = args

    def trigger(self, event) -> bool:
        raise NotImplementedError()

    def describe(self):
        raise NotImplementedError()


class OrFilterGroup(EventFilterBaseGroup):
    """Only one child filter has to match"""

    def trigger(self, event: Any) -> bool:
        for f in self.filters:
            if f.trigger(event):
                return True
        return False

    def describe(self) -> str:
        objs = [f.describe() for f in self.filters]
        return f'({" or ".join(objs)})'


class AndFilterGroup(EventFilterBaseGroup):
    """All child filters have to match"""

    def trigger(self, event: Any) -> bool:
        for f in self.filters:
            if not f.trigger(event):
                return False
        return True

    def describe(self) -> str:
        objs = [f.describe() for f in self.filters]
        return f'({" and ".join(objs)})'



from . import parameters
from .funcs import list_files, sort_files
from .pending_future import PendingFuture
from .single_task import SingleTask
from .rgb_hsv import hsb_to_rgb, rgb_to_hsb
from .exceptions import format_exception



from HABApp.openhab.definitions import OnOffValue
from HABApp.openhab.items.base_item import OpenhabItem
from HABApp.openhab.items.commands import OnOffCommand


class SwitchItem(OpenhabItem, OnOffCommand):
    """SwitchItem which accepts and converts the data types from OpenHAB

    :ivar str name:
    :ivar Tuple[str, ...] value:

    :ivar Optional[str] label:
    :ivar FrozenSet[str] tags:
    :ivar FrozenSet[str] groups:
    :ivar Mapping[str, MetaData] metadata:
    """

    def set_value(self, new_value) -> bool:

        if isinstance(new_value, OnOffValue):
            new_value = new_value.value

        if new_value is not None and new_value != OnOffValue.ON and new_value != OnOffValue.OFF:
            raise ValueError(f'Invalid value for SwitchItem {self.name}: {new_value}')
        return super().set_value(new_value)

    def is_on(self) -> bool:
        """Test value against on-value"""
        return self.value == OnOffValue.ON

    def is_off(self) -> bool:
        """Test value against off-value"""
        return self.value == OnOffValue.OFF

    def __str__(self):
        return self.value

    def __eq__(self, other):
        if isinstance(other, SwitchItem):
            return self.value == other.value
        elif isinstance(other, str):
            return self.value == other
        elif isinstance(other, int):
            if other and self.is_on():
                return True
            if not other and self.is_off():
                return True
            return False

        return NotImplemented

    def __bool__(self):
        return self.is_on()



from typing import Final, Optional, Callable, Any

from HABApp.openhab.connection_logic.plugin_things.cfg_validator import UserItem
from .formatter import TYPE_FORMATTER, EmptyFormatter, ValueFormatter


class BuilderBase:
    def create_formatter(self, item_obj: UserItem) -> 'TYPE_FORMATTER':
        raise NotImplementedError()


class ConstValueFormatterBuilder(BuilderBase):
    def __init__(self, value: str, condition: Optional[Callable] = None):
        self.value: Final = value
        self.condition: Final = condition

    def create_formatter(self, item_obj: UserItem) -> 'TYPE_FORMATTER':
        if self.condition is None or self.condition(item_obj):
            return ValueFormatter(self.value)
        return EmptyFormatter()


class ValueFormatterBuilder(BuilderBase):
    def __init__(self, name: str, fmt_value: str):
        self.name: Final = name
        self.fmt_value: Final = fmt_value

    def get_value(self, item_obj: UserItem):
        return getattr(item_obj, self.name)

    def create_formatter(self, item_obj: UserItem) -> 'TYPE_FORMATTER':
        value = self.get_value(item_obj)
        if not isinstance(value, str):
            raise ValueError('Expected str!')

        value = value.strip()
        if not value:
            return EmptyFormatter()
        return ValueFormatter(self.fmt_value.format(value))


class MultipleValueFormatterBuilder(ValueFormatterBuilder):

    def __init__(self, name: str, fmt_value: str, wrapped_by: str):
        super().__init__(name, fmt_value)
        self.wrapped_by: Final = wrapped_by

    def create_formatter(self, item_obj: UserItem) -> 'TYPE_FORMATTER':
        values = self.get_value(item_obj)
        if not isinstance(values, (list, set, tuple, frozenset)):
            raise ValueError('Expected container!')

        values = map(lambda x: x.strip(), values)

        # remove all empty str and sort
        values = sorted(filter(None, values))
        if not values:
            return EmptyFormatter()

        value = ', '.join(map(self.fmt_value.format, values))
        return ValueFormatter(self.wrapped_by.format(value))


class LinkFormatter(BuilderBase):

    def create_formatter(self, item_obj: UserItem) -> 'TYPE_FORMATTER':
        link = item_obj.link.strip()
        if not link:
            return EmptyFormatter()

        value = f'channel = "{link:s}"'
        if item_obj.metadata:
            value += ','

        return ValueFormatter(value)


def metadata_key_value(key: str, value: Any):
    return f'{key}={value}' if not isinstance(value, str) else f'{key}="{value}"'


class MetadataFormatter(BuilderBase):

    def create_formatter(self, item_obj: UserItem) -> 'TYPE_FORMATTER':
        metdata = item_obj.metadata
        if not metdata:
            return EmptyFormatter()

        strs = []
        for name, __meta in sorted(metdata.items(), key=lambda x: x[0]):
            value = __meta['value']
            config = __meta['config']

            # a=1 or a="test"
            metadata_str = metadata_key_value(name, value)
            if config:
                config_strs = []
                for config_key, config_value in sorted(config.items(), key=lambda x: x[0]):
                    config_strs.append(metadata_key_value(config_key, config_value))

                metadata_str += f' [{", ".join(config_strs)}]'

            strs.append(metadata_str)

        return ValueFormatter(', '.join(strs))



from pathlib import Path

import HABApp
from HABApp.rule.rule_hook import HABAppRuleHook
from HABApp.rule_manager import RuleFile
from .bench_habapp import HABAppBenchRule
from .bench_mqtt import MqttBenchRule
from .bench_oh import OpenhabBenchRule


class BenchFile(RuleFile):
    def __init__(self, rule_manager):
        super().__init__(rule_manager, 'BenchmarkFile', path=Path('BenchmarkFile'))

    def create_rules(self, created_rules: list):
        HABAppRuleHook.in_dict(globals(), created_rules.append, self.suggest_rule_name, self.rule_manager.runtime, self)

        rule_ha = rule = HABAppBenchRule()
        if HABApp.CONFIG.mqtt.connection.host:
            rule = rule.link_rule(MqttBenchRule())
        if HABApp.CONFIG.openhab.connection.url:
            rule = rule.link_rule(OpenhabBenchRule())

        rule_ha.run.at(5, rule_ha.do_bench_start)



from HABApp.core.errors import HABAppException


class HABAppOpenhabError(HABAppException):
    pass


# ----------------------------------------------------------------------------------------------------------------------
# Connection errors
# ----------------------------------------------------------------------------------------------------------------------
class OpenhabConnectionNotSetUpError(HABAppOpenhabError):
    pass


class OpenhabDisconnectedError(HABAppOpenhabError):
    pass


class ExpectedSuccessFromOpenhab(HABAppOpenhabError):
    pass


# ----------------------------------------------------------------------------------------------------------------------
# OpenHAB errors
# ----------------------------------------------------------------------------------------------------------------------
class SendCommandNotSupported(HABAppOpenhabError):
    pass


class ItemNotFoundError(HABAppOpenhabError):

    @classmethod
    def from_name(cls, name: str):
        return cls(f'Item "{name}" not found!')


class ItemNotEditableError(HABAppOpenhabError):

    @classmethod
    def from_name(cls, name: str):
        return cls(f'Item "{name}" is not editable!')


class ThingNotFoundError(HABAppOpenhabError):

    @classmethod
    def from_uid(cls, uid: str):
        return cls(f'Thing "{uid}" not found!')


class ThingNotEditableError(HABAppOpenhabError):

    @classmethod
    def from_uid(cls, uid: str):
        return cls(f'Thing "{uid}" is not editable!')


# ----------------------------------------------------------------------------------------------------------------------
# RestAPI Errors
# ----------------------------------------------------------------------------------------------------------------------
class MetadataNotEditableError(HABAppOpenhabError):

    @classmethod
    def create_text(cls, item: str, namespace: str):
        return cls(f'Metadata {namespace} for {item} is not editable!')



from HABApp.openhab.definitions import OnOffValue, UpDownValue
from HABApp.openhab.interface import send_command


class OnOffCommand:

    def is_on(self) -> bool:
        """Test value against on-value"""
        raise NotImplementedError()

    def is_off(self) -> bool:
        """Test value against off-value"""
        raise NotImplementedError()

    def on(self):
        """Command item on"""
        send_command(self, OnOffValue.ON)

    def off(self):
        """Command item off"""
        send_command(self, OnOffValue.OFF)


class PercentCommand:
    def percent(self, value: float):
        """Command to value (in percent)"""
        assert 0 <= value <= 100, value
        send_command(self, str(value))


class UpDownCommand:
    def up(self):
        """Command up"""
        send_command(self, UpDownValue.UP)

    def down(self):
        """Command down"""
        send_command(self, UpDownValue.DOWN)

    def is_up(self) -> bool:
        """Test value against on-value"""
        raise NotImplementedError()

    def is_down(self) -> bool:
        """Test value against off-value"""
        raise NotImplementedError()



from typing import Tuple

from HABApp.core.events import ComplexEventValue
from HABApp.openhab.items.base_item import OpenhabItem
from HABApp.openhab.item_to_reg import get_members


class GroupItem(OpenhabItem):
    """GroupItem which accepts and converts the data types from OpenHAB

    :ivar str name:
    :ivar str value:

    :ivar Optional[str] label:
    :ivar FrozenSet[str] tags:
    :ivar FrozenSet[str] groups:
    :ivar Mapping[str, MetaData] metadata:
    """

    def set_value(self, new_value) -> bool:

        if isinstance(new_value, ComplexEventValue):
            new_value = new_value.value
        return super().set_value(new_value)

    @property
    def members(self) -> Tuple[OpenhabItem, ...]:
        """Resolves and then returns all group members"""

        return get_members(self.name)



import asyncio
import logging
import typing

import paho.mqtt.client as mqtt

import HABApp
from HABApp.core.asyncio import async_context
from HABApp.core.const.topics import TOPIC_EVENTS
from HABApp.core.errors import ItemNotFoundException
from HABApp.core.internals import uses_post_event, uses_get_item, uses_item_registry
from HABApp.core.wrapper import log_exception
from HABApp.mqtt.events import MqttValueChangeEvent, MqttValueUpdateEvent
from HABApp.mqtt.mqtt_payload import get_msg_payload
from HABApp.runtime import shutdown

log = logging.getLogger('HABApp.mqtt.connection')
log_msg = logging.getLogger(f'{TOPIC_EVENTS}.mqtt')


post_event = uses_post_event()
get_item = uses_get_item()
Items = uses_item_registry()


class MqttStatus:
    def __init__(self):
        self.loop_started = False
        self.connected = False
        self.client: typing.Optional[mqtt.Client] = None
        self.subscriptions: typing.List[typing.Tuple[str, int]] = []


STATUS = MqttStatus()


def setup():
    config = HABApp.config.CONFIG.mqtt

    # config changes
    config.subscribe.subscribe_for_changes(subscription_changed)
    config.connection.subscribe_for_changes(connect)

    # shutdown
    shutdown.register_func(disconnect, msg='Disconnecting MQTT')


def connect():
    config = HABApp.config.CONFIG.mqtt

    if not config.connection.host:
        log.info('MQTT disabled')
        disconnect()
        return None

    if STATUS.connected:
        log.info('disconnecting')
        STATUS.client.disconnect()
        STATUS.connected = False

    STATUS.client = mqtt_client = mqtt.Client(
        client_id=config.connection.client_id,
        clean_session=False
    )

    if config.connection.tls.enabled:
        log.debug("TLS enabled")
        # add option to specify tls certificate
        ca_cert = config.connection.tls.ca_cert
        if ca_cert is not None and ca_cert.name:
            if not ca_cert.is_file():
                log.error(f'Ca cert file does not exist: {ca_cert}')
                # don't connect without the properly set certificate
                disconnect()
                return None
            else:
                log.debug(f"CA cert path: {ca_cert}")
                mqtt_client.tls_set(ca_cert)
        else:
            mqtt_client.tls_set()

        # we can only set tls_insecure if we have a tls connection
        if config.connection.tls.insecure:
            log.warning('Verification of server hostname in server certificate disabled!')
            log.warning('Use this only for testing, not for a real system!')
            mqtt_client.tls_insecure_set(True)

    # set user/pw if required
    user = config.connection.user
    pw = config.connection.password
    if user:
        mqtt_client.username_pw_set(user, pw if pw else None)

    # setup callbacks
    mqtt_client.on_connect = on_connect
    mqtt_client.on_disconnect = on_disconnect
    mqtt_client.on_message = process_msg

    mqtt_client.connect_async(
        config.connection.host, port=config.connection.port, keepalive=60
    )

    log.info(f'Connecting to {config.connection.host}:{config.connection.port}')

    if not STATUS.loop_started:
        mqtt_client.loop_start()
    STATUS.loop_started = True


@log_exception
def disconnect():
    if STATUS.connected:
        STATUS.client.disconnect()
        STATUS.connected = False
    if STATUS.loop_started:
        STATUS.client.loop_stop()
        STATUS.loop_started = False

    STATUS.client = None


@log_exception
def on_connect(client, userdata, flags, rc):
    log.log(logging.INFO if not rc else logging.ERROR, mqtt.connack_string(rc))
    if rc:
        return None
    STATUS.connected = True

    STATUS.subscriptions.clear()
    subscription_changed()


@log_exception
def on_disconnect(client, userdata, rc):
    log.log(logging.INFO if not rc else logging.ERROR, f'Disconnect: {mqtt.error_string(rc)} ({rc})')
    STATUS.connected = False


@log_exception
def subscription_changed():
    if not STATUS.connected:
        return None

    if STATUS.subscriptions:
        unsubscribe = [k[0] for k in STATUS.subscriptions]
        log.debug('Unsubscribing from:')
        for t in unsubscribe:
            log.debug(f' - "{t}"')
        STATUS.client.unsubscribe(unsubscribe)

    topics = HABApp.config.CONFIG.mqtt.subscribe.topics
    default_qos = HABApp.config.CONFIG.mqtt.subscribe.qos
    STATUS.subscriptions = [(topic, qos if qos is not None else default_qos) for topic, qos in topics]
    log.debug('Subscribing to:')
    for topic, qos in STATUS.subscriptions:
        log.debug(f' - "{topic}" (QoS {qos:d})')
    STATUS.client.subscribe(STATUS.subscriptions)


@log_exception
def process_msg(client, userdata, message: mqtt.MQTTMessage):

    topic, payload = get_msg_payload(message)
    if topic is None:
        return None

    # Post events
    asyncio.run_coroutine_threadsafe(send_event_async(topic, payload, message.retain), HABApp.core.const.loop)


async def send_event_async(topic, payload, retain: bool):
    async_context.set('MQTT')

    _item = None    # type: typing.Optional[HABApp.mqtt.items.MqttBaseItem]
    try:
        _item = get_item(topic)   # type: HABApp.mqtt.items.MqttBaseItem
    except ItemNotFoundException:
        # only create items for if the message has the retain flag
        if retain:
            _item = Items.add_item(HABApp.mqtt.items.MqttItem(topic))

    # we don't have an item -> we process only the event
    if _item is None:
        post_event(topic, MqttValueUpdateEvent(topic, payload))
        return None

    # Remember state and update item before doing callbacks
    _old_state = _item.value
    _item.set_value(payload)

    post_event(topic, MqttValueUpdateEvent(topic, payload))
    if payload != _old_state:
        post_event(topic, MqttValueChangeEvent(topic, payload, _old_state))



from typing import Any

from HABApp.openhab.items.base_item import OpenhabItem
from ..definitions import OpenClosedValue
from ...core.const import MISSING
from ..errors import SendCommandNotSupported


class ContactItem(OpenhabItem):
    """ContactItem

    :ivar str name:
    :ivar str value:

    :ivar Optional[str] label:
    :ivar FrozenSet[str] tags:
    :ivar FrozenSet[str] groups:
    :ivar Mapping[str, MetaData] metadata:
    """

    def set_value(self, new_value) -> bool:

        if isinstance(new_value, OpenClosedValue):
            new_value = new_value.value

        if new_value is not None and new_value != OpenClosedValue.OPEN and new_value != OpenClosedValue.CLOSED:
            raise ValueError(f'Invalid value for ContactItem: {new_value}')
        return super().set_value(new_value)

    def is_open(self) -> bool:
        """Test value against open-value"""
        return self.value == OpenClosedValue.OPEN

    def is_closed(self) -> bool:
        """Test value against closed-value"""
        return self.value == OpenClosedValue.CLOSED

    def oh_send_command(self, value: Any = MISSING):
        raise SendCommandNotSupported(f'{self.__class__.__name__} does not support send command! '
                                      'See openHAB documentation for details.')

    def __str__(self):
        return self.value

    def __eq__(self, other):
        if isinstance(other, ContactItem):
            return self.value == other.value
        elif isinstance(other, str):
            return self.value == other
        elif isinstance(other, int):
            if other and self.is_open():
                return True
            if not other and self.is_closed():
                return True
            return False

        return NotImplemented



import asyncio
import time
from pathlib import Path
from typing import Dict, Set, Optional, List, Any

import HABApp
from HABApp.core.files.file import HABAppFile
from HABApp.core.files.folders import add_folder as add_habapp_folder
from HABApp.core.files.watcher import AggregatingAsyncEventHandler
from HABApp.core.lib import PendingFuture
from HABApp.core.logger import log_warning, HABAppError
from HABApp.openhab.connection_handler.func_async import async_get_things
from HABApp.openhab.connection_logic.plugin_things.cfg_validator import validate_cfg, InvalidItemNameError
from HABApp.openhab.connection_logic.plugin_things.filters import THING_ALIAS, CHANNEL_ALIAS
from HABApp.openhab.connection_logic.plugin_things.filters import apply_filters, log_overview
from ._log import log
from .item_worker import create_item, cleanup_items
from .file_writer import ItemsFileWriter
from .thing_worker import update_thing_cfg
from .._plugin import OnConnectPlugin


class DuplicateItemError(Exception):
    pass


class ManualThingConfig(OnConnectPlugin):

    def __init__(self):
        super().__init__()
        self.created_items: Dict[str, Set[str]] = {}
        self.do_cleanup = PendingFuture(self.clean_items, 120)

        self.watcher: Optional[AggregatingAsyncEventHandler] = None

        self.cache_ts: float = 0.0
        self.cache_cfg: List[Dict[str, Any]] = []

    def setup(self):
        path = HABApp.CONFIG.directories.config
        if path is None:
            return None

        class HABAppThingConfigFile(HABAppFile):
            LOGGER = log
            LOAD_FUNC = self.file_load
            UNLOAD_FUNC = self.file_unload

        folder = add_habapp_folder('config/', path, 50)
        folder.add_file_type(HABAppThingConfigFile)
        self.watcher = folder.add_watch('.yml')

    async def file_unload(self, prefix: str, path: Path):
        return None

    async def on_connect_function(self):
        if self.watcher is None:
            return None

        await asyncio.sleep(0.3)
        await self.load_thing_data(always=True)
        await self.watcher.trigger_all()

    @HABApp.core.wrapper.ignore_exception
    async def clean_items(self):
        items = set()
        for s in self.created_items.values():
            items.update(s)
        await cleanup_items(items)

    async def load_thing_data(self, always: bool) -> List[Dict[str, Any]]:
        if always or not self.cache_cfg or time.time() - self.cache_ts > 20:
            self.cache_cfg = [k.dict(by_alias=True) for k in await async_get_things()]
            self.cache_ts = time.time()
        return self.cache_cfg

    async def file_load(self, name: str, path: Path):
        # we have to check the naming structure because we get file events for the whole folder
        _name = path.name.lower()
        if not _name.startswith('thing_') or not _name.endswith('.yml'):
            log.warning(f'Name for "{name}" does not start with "thing_" -> skip!')
            return None

        # only load if we don't supply the data
        data = await self.load_thing_data(always=False)

        # remove created items
        self.created_items.pop(path.name, None)
        created_items = self.created_items.setdefault(path.name, set())

        # shedule cleanup
        self.do_cleanup.reset()

        # output file
        items_file_path = path.with_suffix('.items')
        items_file_writer = ItemsFileWriter()

        # we also get events when the file gets deleted
        if not path.is_file():
            log.debug(f'File {path} does not exist -> skipping Thing configuration!')
            return None
        log.debug(f'Loading {name}!')

        # load the config file
        with path.open(mode='r', encoding='utf-8') as file:
            try:
                cfg = HABApp.core.const.yml.load(file)
            except Exception as e:
                HABAppError(log).add_exception(e).dump()
                return None

        # validate configuration
        cfg = validate_cfg(cfg, path.name)
        if cfg is None:
            return None

        # if one entry has test set we show an overview of all the things
        if any(map(lambda x: x.test, cfg)):
            log_overview(data, THING_ALIAS, 'Thing overview')

        # process each thing part in the cfg
        for cfg_entry in cfg:
            test: bool = cfg_entry.test
            things = list(apply_filters(cfg_entry.filter, data, test))

            # show a warning we found no Things
            if not things:
                log_warning(log, f'No things matched for {cfg_entry.filter}')
                continue

            # update thing configuration
            if cfg_entry.thing_config:
                await update_thing_cfg(cfg_entry.thing_config, things, test)

            try:
                # item creation for every thing
                create_items = {}
                shown_types = set()
                for thing in things:
                    thing_context = {k: thing.get(alias, '') for k, alias in THING_ALIAS.items()}

                    # create items without channel
                    for item_cfg in cfg_entry.get_items(thing_context):
                        name = item_cfg.name
                        if name in create_items:
                            raise DuplicateItemError(f'Duplicate item: {name}')
                        create_items[name] = item_cfg

                    # Channel overview, only if we have something configured
                    if test and cfg_entry.channels:
                        __thing_type = thing_context['thing_type']
                        if __thing_type not in shown_types:
                            shown_types.add(__thing_type)
                            log_overview(thing['channels'], CHANNEL_ALIAS, heading=f'Channels for {__thing_type}')

                    # do channel things
                    for channel_cfg in cfg_entry.channels:
                        channels = apply_filters(channel_cfg.filter, thing['channels'], test)
                        for channel in channels:
                            channel_context = {k: channel.get(alias, '') for k, alias in CHANNEL_ALIAS.items()}
                            channel_context.update(thing_context)

                            for item_cfg in channel_cfg.get_items(channel_context):
                                item_cfg.link = channel['uid']
                                name = item_cfg.name

                                if name in create_items:
                                    raise DuplicateItemError(f'Duplicate item: {name}')
                                create_items[name] = item_cfg

                    # newline only if we create logs
                    if test and (cfg_entry.create_items or cfg_entry.channels):
                        log.info('')
            except InvalidItemNameError as e:
                HABAppError(log).add_exception(e).dump()
                continue
            except DuplicateItemError as e:
                # Duplicates should never happen, the user clearly made a mistake, that's why we exit here
                HABAppError(log).add_exception(e).dump()
                return None

            # Create all items
            for item_cfg in create_items.values():
                created = await create_item(item_cfg, test)
                if created:
                    created_items.add(item_cfg.name)

            self.do_cleanup.reset()

            items_file_writer.add_items(create_items.values())
            self.cache_cfg = []

        items_file_writer.create_file(items_file_path)


PLUGIN_MANUAL_THING_CFG = ManualThingConfig.create_plugin()



import HABApp


class MqttValueUpdateEvent(HABApp.core.events.ValueUpdateEvent):
    # Copy the annotations, otherwise the code won't work from py3.10 on
    __annotations__ = HABApp.core.events.ValueUpdateEvent.__annotations__.copy()


class MqttValueChangeEvent(HABApp.core.events.ValueChangeEvent):
    # Copy the annotations, otherwise the code won't work from py3.10 on
    __annotations__ = HABApp.core.events.ValueChangeEvent.__annotations__.copy()



import logging
from typing import Union

import HABApp
from HABApp.core.const.topics import TOPIC_FILES as T_FILES
from HABApp.core.events.habapp_events import RequestFileUnloadEvent, RequestFileLoadEvent
from HABApp.core.events import EventFilter
from HABApp.core.internals import wrap_func, EventBusListener, uses_event_bus

log = logging.getLogger('HABApp.Files')
event_bus = uses_event_bus()


async def _process_event(event: Union[RequestFileUnloadEvent, RequestFileLoadEvent]):
    name = event.name
    await HABApp.core.files.manager.process_file(name, HABApp.core.files.folders.get_path(name))


async def setup_file_manager():
    # Setup events so we can process load/unload
    event_bus.add_listener(
        EventBusListener(
            T_FILES, wrap_func(_process_event), EventFilter(RequestFileUnloadEvent)
        )
    )
    event_bus.add_listener(
        EventBusListener(
            T_FILES, wrap_func(_process_event), EventFilter(RequestFileLoadEvent)
        )
    )



import logging

import HABApp
from HABApp.core.wrapper import ignore_exception
from HABApp.openhab.item_to_reg import add_to_registry, fresh_item_sync, remove_from_registry, \
    remove_thing_from_registry, add_thing_to_registry
from HABApp.openhab.map_items import map_item
from ._plugin import OnConnectPlugin
from ..interface_async import async_get_items, async_get_things
from ...core.internals import uses_item_registry

log = logging.getLogger('HABApp.openhab.items')

Items = uses_item_registry()


class LoadAllOpenhabItems(OnConnectPlugin):

    @ignore_exception
    async def on_connect_function(self):
        data = await async_get_items(all_metadata=True)
        if data is None:
            return None

        fresh_item_sync()

        found_items = len(data)
        for _dict in data:
            item_name = _dict['name']
            new_item = map_item(item_name, _dict['type'], _dict['state'], _dict.get('label'),
                                frozenset(_dict['tags']), frozenset(_dict['groupNames']),
                                _dict.get('metadata', {}))   # type: HABApp.openhab.items.OpenhabItem
            if new_item is None:
                continue
            add_to_registry(new_item, True)

        # remove items which are no longer available
        ist = set(Items.get_item_names())
        soll = {k['name'] for k in data}
        for k in ist - soll:
            if isinstance(Items.get_item(k), HABApp.openhab.items.OpenhabItem):
                remove_from_registry(k)

        log.info(f'Updated {found_items:d} Items')

        # try to update things, too
        data = await async_get_things()

        Thing = HABApp.openhab.items.Thing
        for thing_cfg in data:
            add_thing_to_registry(thing_cfg)

        # remove things which were deleted
        ist = set(Items.get_item_names())
        soll = {k.uid for k in data}
        for k in ist - soll:
            if isinstance(Items.get_item(k), Thing):
                remove_thing_from_registry(k)

        log.info(f'Updated {len(data):d} Things')
        return None


PLUGIN_LOAD_ITEMS = LoadAllOpenhabItems.create_plugin()



from colorsys import hsv_to_rgb as _hsv_to_rgb
from colorsys import rgb_to_hsv as _rgb_to_hsv
from typing import Optional, Tuple, Union


def rgb_to_hsb(r: Union[int, float], g: Union[int, float], b: Union[int, float],
               max_rgb_value: int = 255, ndigits: Optional[int] = 2) -> Tuple[float, float, float]:
    """Convert from rgb to hsb/hsv

    :param r: red value
    :param g: green value
    :param b: blue value
    :param max_rgb_value: maximal possible rgb value (e.g. 255 for 8 bit or 65.535 for 16bit values)
    :param ndigits: Round the hsb values to the specified digits, None to disable rounding
    :return: Values for hue, saturation and brightness / value
    """
    h, s, v = _rgb_to_hsv(r / max_rgb_value, g / max_rgb_value, b / max_rgb_value)
    h *= 360
    s *= 100
    v *= 100

    if ndigits is not None:
        h = round(h, ndigits)
        s = round(s, ndigits)
        v = round(v, ndigits)

    return h, s, v


def hsb_to_rgb(h, s, b, max_rgb_value=255) -> Tuple[int, int, int]:
    """Convert from rgb to hsv/hsb

    :param h: hue
    :param s: saturation
    :param b: brightness / value
    :param max_rgb_value: maximal value for the returned rgb values (e.g. 255 for 8 bit or 65.535 16bit values)
    :return: Values for red, green and blue
    """
    r, g, b = _hsv_to_rgb(h / 360, s / 100, b / 100)
    return round(r * max_rgb_value), round(g * max_rgb_value), round(b * max_rgb_value)



import random
from datetime import datetime as dt_datetime, time as dt_time, timedelta as dt_timedelta
from typing import Iterable, Union, Callable, Any

import HABApp.rule_ctx
from HABApp.core.const.const import PYTHON_310
from HABApp.core.internals import ContextProvidingObj, HINT_CONTEXT_OBJ
from HABApp.core.internals import wrap_func
from HABApp.rule.scheduler.executor import WrappedFunctionExecutor
from HABApp.rule.scheduler.scheduler import HABAppScheduler as _HABAppScheduler
from eascheduler import SchedulerView
from eascheduler.jobs import CountdownJob, DawnJob, DayOfWeekJob, DuskJob, OneTimeJob, ReoccurringJob, SunriseJob, \
    SunsetJob

if PYTHON_310:
    from typing import TypeAlias, ParamSpec
else:
    from typing_extensions import TypeAlias, ParamSpec


HINT_CB_P = ParamSpec('HINT_CB_P')
HINT_CB: TypeAlias = Callable[HINT_CB_P, Any]


class HABAppSchedulerView(SchedulerView, ContextProvidingObj):
    def __init__(self, context: 'HABApp.rule_ctx.HABAppRuleContext'):
        super().__init__(_HABAppScheduler(), WrappedFunctionExecutor)
        self._habapp_rule_ctx: HINT_CONTEXT_OBJ = context

    def at(self, time: Union[None, dt_datetime, dt_timedelta, dt_time, int],
           callback: HINT_CB, *args: HINT_CB_P.args, **kwargs: HINT_CB_P.kwargs) -> OneTimeJob:
        callback = wrap_func(callback, context=self._habapp_rule_ctx)
        return super().at(time, callback, *args, **kwargs)

    def countdown(self, expire_time: Union[dt_timedelta, float, int],
                  callback: HINT_CB, *args: HINT_CB_P.args, **kwargs: HINT_CB_P.kwargs) -> CountdownJob:
        callback = wrap_func(callback, context=self._habapp_rule_ctx)
        return super().countdown(expire_time, callback, *args, **kwargs)

    def every(self, start_time: Union[None, dt_datetime, dt_timedelta, dt_time, int],
              interval: Union[int, float, dt_timedelta],
              callback: HINT_CB, *args: HINT_CB_P.args, **kwargs: HINT_CB_P.kwargs) -> ReoccurringJob:
        callback = wrap_func(callback, context=self._habapp_rule_ctx)
        return super().every(start_time, interval, callback, *args, **kwargs)

    def on_day_of_week(self, time: Union[dt_time, dt_datetime], weekdays: Union[str, Iterable[Union[str, int]]],
                       callback: HINT_CB, *args: HINT_CB_P.args, **kwargs: HINT_CB_P.kwargs) -> DayOfWeekJob:
        callback = wrap_func(callback, context=self._habapp_rule_ctx)
        return super().on_day_of_week(time, weekdays, callback, *args, **kwargs)

    def on_every_day(self, time: Union[dt_time, dt_datetime],
                     callback: HINT_CB, *args: HINT_CB_P.args, **kwargs: HINT_CB_P.kwargs) -> DayOfWeekJob:
        """Create a job that will run at a certain time of day

        :param time: Time when the job will run
        :param callback: |param_scheduled_cb|
        :param args: |param_scheduled_cb_args|
        :param kwargs: |param_scheduled_cb_kwargs|
        """
        callback = wrap_func(callback, context=self._habapp_rule_ctx)
        return super().on_day_of_week(time, 'all', callback, *args, **kwargs)

    def on_sunrise(self, callback: HINT_CB, *args: HINT_CB_P.args, **kwargs: HINT_CB_P.kwargs) -> SunriseJob:
        callback = wrap_func(callback, context=self._habapp_rule_ctx)
        return super().on_sunrise(callback, *args, **kwargs)

    def on_sunset(self, callback: HINT_CB, *args: HINT_CB_P.args, **kwargs: HINT_CB_P.kwargs) -> SunsetJob:
        callback = wrap_func(callback, context=self._habapp_rule_ctx)
        return super().on_sunset(callback, *args, **kwargs)

    def on_sun_dawn(self, callback: HINT_CB, *args: HINT_CB_P.args, **kwargs: HINT_CB_P.kwargs) -> DawnJob:
        callback = wrap_func(callback, context=self._habapp_rule_ctx)
        return super().on_sun_dawn(callback, *args, **kwargs)

    def on_sun_dusk(self, callback: HINT_CB, *args: HINT_CB_P.args, **kwargs: HINT_CB_P.kwargs) -> DuskJob:
        callback = wrap_func(callback, context=self._habapp_rule_ctx)
        return super().on_sun_dusk(callback, *args, **kwargs)

    def soon(self, callback: HINT_CB, *args: HINT_CB_P.args, **kwargs: HINT_CB_P.kwargs) -> OneTimeJob:
        """
        Run the callback as soon as possible.

        :param callback: |param_scheduled_cb|
        :param args: |param_scheduled_cb_args|
        :param kwargs: |param_scheduled_cb_kwargs|
        """
        return self.at(None, callback, *args, **kwargs)

    def every_minute(self, callback: HINT_CB, *args: HINT_CB_P.args, **kwargs: HINT_CB_P.kwargs) -> ReoccurringJob:
        """Picks a random second and runs the callback every minute

        :param callback: |param_scheduled_cb|
        :param args: |param_scheduled_cb_args|
        :param kwargs: |param_scheduled_cb_kwargs|
        """
        start = dt_timedelta(seconds=random.randint(0, 60 - 1))
        interval = dt_timedelta(seconds=60)
        return self.every(start, interval, callback, *args, **kwargs)

    def every_hour(self, callback: HINT_CB, *args: HINT_CB_P.args, **kwargs: HINT_CB_P.kwargs) -> ReoccurringJob:
        """Picks a random minute and second and run the callback every hour

        :param callback: |param_scheduled_cb|
        :param args: |param_scheduled_cb_args|
        :param kwargs: |param_scheduled_cb_kwargs|
        """
        start = dt_timedelta(seconds=random.randint(0, 3600 - 1))
        interval = dt_timedelta(hours=1)
        return self.every(start, interval, callback, *args, **kwargs)



from .base_event import OpenhabEvent


class ChannelTriggeredEvent(OpenhabEvent):
    """
    :ivar str name:
    :ivar str event:
    :ivar str channel:
    """
    name: str
    event: str
    channel: str

    def __init__(self, name: str = '', event: str = '', channel: str = ''):
        super().__init__()

        self.name: str = name
        self.event: str = event
        self.channel: str = channel

    @classmethod
    def from_dict(cls, topic: str, payload: dict):
        return cls(topic[17:-10], payload['event'], payload['channel'])

    def __repr__(self):
        return f'<{self.__class__.__name__} name: {self.name}, event: {self.event}>'


class ChannelDescriptionChangedEvent(OpenhabEvent):
    """
    :ivar str name:
    :ivar str field:
    :ivar str value:
    """
    name: str
    field: str
    value: str

    def __init__(self, name: str = '', field: str = '', value: str = ''):
        super().__init__()

        self.name: str = name
        self.field: str = field
        self.value: str = value

    @classmethod
    def from_dict(cls, topic: str, payload: dict):
        return cls(topic[17:-19], payload['field'], payload['value'])

    def __repr__(self):
        return f'<{self.__class__.__name__} name: {self.name}, field: {self.field}>'



from HABApp.core.internals import EventFilterBase


class NoEventFilter(EventFilterBase):
    """Triggers on all events"""

    def trigger(self, event) -> bool:
        return True

    def describe(self) -> str:
        return f'{self.__class__.__name__}()'



import asyncio
import collections
import time
import typing
from datetime import timedelta

import HABApp
from HABApp.core.errors import ItemNotFoundException
from HABApp.core.internals import HINT_EVENT_BUS_LISTENER, wrap_func
from HABApp.core.wrapper import process_exception
from HABApp.core.internals import uses_item_registry, uses_get_item, uses_event_bus, EventBusListener
from HABApp.core.items import BaseValueItem
from HABApp.core.events import EventFilter, ValueChangeEvent, ValueUpdateEvent


get_item = uses_get_item()
item_registry = uses_item_registry()
event_bus = uses_event_bus()


class AggregationItem(BaseValueItem):

    @classmethod
    def get_create_item(cls, name: str):
        """Creates a new AggregationItem in HABApp and returns it or returns the
        already existing item with the given name

        :param name: item name
        :return: item
        """
        assert isinstance(name, str), type(name)

        try:
            item = get_item(name)
        except ItemNotFoundException:
            item = cls(name)
            item_registry.add_item(item)

        assert isinstance(item, cls), f'{cls} != {type(item)}'
        return item

    def __init__(self, name: str):
        super().__init__(name)
        self.__period: float = 0
        self.__aggregation_func: typing.Callable[[typing.Iterable], typing.Any] = lambda x: x

        self._ts: typing.Deque[float] = collections.deque()
        self._vals: typing.Deque[typing.Any] = collections.deque()

        self.__listener: typing.Optional[HINT_EVENT_BUS_LISTENER] = None

        self.__task: typing.Optional[asyncio.Future] = None

    def aggregation_func(self, func: typing.Callable[[typing.Iterable], typing.Any]) -> 'AggregationItem':
        """Set the function which will be used to aggregate all values. E.g. ``min`` or ``max``

        :param func: The function which takes an iterator an returns an aggregated value.
                     Important: the function must be **non blocking**!
        """
        self.__aggregation_func = func
        return self

    def aggregation_period(self, period: typing.Union[float, int, timedelta]) -> 'AggregationItem':
        """Set the period in which the items will be aggregated

        :param period: period in seconds
        """
        if isinstance(period, timedelta):
            period = period.total_seconds()

        assert period > 0, period
        self.__period = period

        # Clean old items (e.g. if we made the period shorter)
        while len(self._ts) > 1 and self._ts[1] + self.__period < time.time():
            self._ts.popleft()
            self._vals.popleft()

        return self

    def aggregation_source(self, source: typing.Union[BaseValueItem, str],
                           only_changes: bool = False) -> 'AggregationItem':
        """Set the source item which changes will be aggregated

        :param source: name or Item obj
        :param only_changes: if true only value changes instead of value updates will be added
        """

        # If we already have one we cancel it
        if self.__listener is not None:
            self.__listener.cancel()
            self.__listener = None

        self.__listener = EventBusListener(
            topic=source.name if isinstance(source, HABApp.core.items.BaseValueItem) else source,
            callback=wrap_func(self._add_value, name=f'{self.name}.add_value'),
            event_filter=EventFilter(ValueChangeEvent if only_changes else ValueUpdateEvent)
        )
        event_bus.add_listener(self.__listener)
        return self

    def _on_item_removed(self):
        super()._on_item_removed()

        if self.__listener is not None:
            self.__listener.cancel()
            self.__listener = None

        if self.__task is not None:
            self.__task.cancel()
            self.__task = None

    async def __update_task(self):
        try:
            while len(self._ts) > 1:
                ts = self._ts[1]
                now = time.time()

                left = (ts + self.__period) - now
                while left > 0:
                    await asyncio.sleep(left)
                    # sometimes we wake up to early
                    now = time.time()
                    left = (ts + self.__period) - now

                self._ts.popleft()
                self._vals.popleft()

                # old entries are removed -> now do the aggregation
                try:
                    val = self.__aggregation_func(self._vals)
                except Exception as e:
                    process_exception(self.__aggregation_func, e)
                    continue
                self.post_value(val)

        except Exception as e:
            process_exception(self.__update_task, e)
        finally:
            self.__task = None
        return None

    async def _add_value(self, event: ValueChangeEvent):
        self._ts.append(time.time())
        self._vals.append(event.value)

        if self.__task is None:
            self.__task = asyncio.create_task(self.__update_task())

        try:
            val = self.__aggregation_func(self._vals)
        except Exception as e:
            process_exception(self.__aggregation_func, e)
            return None

        self.post_value(val)
        return None



from typing import Any, Callable, Optional, Union

from HABApp.core.events import EventFilter
from HABApp.core.internals import EventBusListener
from HABApp.core.items import HINT_ITEM_OBJ


class ListenerCreatorBase:
    def __init__(self, item: HINT_ITEM_OBJ, callback: Callable[[Any], Any]):
        self.item = item
        self.callback = callback

        self.listener: Optional[EventBusListener] = None
        self.active = True

    def create_listener(self) -> EventBusListener:
        raise NotImplementedError()

    def listen(self):
        if not self.active:
            return None

        if self.listener is None:
            self.listener = self.create_listener()

    def cancel(self):
        if not self.active:
            return None

        if self.listener is not None:
            self.listener.cancel()
            self.listener = None


class EventListenerCreator(ListenerCreatorBase):
    def __init__(self, item: HINT_ITEM_OBJ, callback: Callable[[Any], Any], event_filter: EventFilter):
        super(EventListenerCreator, self).__init__(item, callback)
        self.event_filter = event_filter

    def create_listener(self) -> EventBusListener:
        return self.item.listen_event(self.callback, self.event_filter)


class NoUpdateEventListenerCreator(ListenerCreatorBase):
    def __init__(self, item: HINT_ITEM_OBJ, callback: Callable[[Any], Any], secs: Union[int, float]):
        super(NoUpdateEventListenerCreator, self).__init__(item, callback)
        self.secs = secs

    def create_listener(self) -> EventBusListener:
        return self.item.watch_update(self.secs).listen_event(self.callback)


class NoChangeEventListenerCreator(ListenerCreatorBase):
    def __init__(self, item: HINT_ITEM_OBJ, callback: Callable[[Any], Any], secs: Union[int, float]):
        super(NoChangeEventListenerCreator, self).__init__(item, callback)
        self.secs = secs

    def create_listener(self) -> EventBusListener:
        return self.item.watch_change(self.secs).listen_event(self.callback)



import asyncio
import typing


class FinishedProcessInfo:
    """Information about the finished process."""

    def __init__(self, returncode: int, stdout: typing.Optional[str], stderr: typing.Optional[str]):
        self.returncode: int = returncode
        self.stdout: typing.Optional[str] = stdout
        self.stderr: typing.Optional[str] = stderr

    def __repr__(self):
        return f'<ProcessInfo: returncode: {self.returncode}, stdout: {self.stdout}, stderr: {self.stderr}>'


async def async_subprocess_exec(callback, program: str, *args, capture_output=True):
    assert isinstance(program, str), type(program)

    proc = None
    stdout = None
    stderr = None

    try:
        proc = await asyncio.create_subprocess_exec(
            program,
            *args,
            stdout=asyncio.subprocess.PIPE if capture_output else None,
            stderr=asyncio.subprocess.PIPE if capture_output else None
        )

        b_stdout, b_stderr = await proc.communicate()
        ret_code = proc.returncode
        if capture_output:
            stdout = b_stdout.decode()
            stderr = b_stderr.decode()
    except asyncio.CancelledError:
        if proc is not None:
            proc.terminate()

        stdout = None
        stderr = None
        ret_code = -2
    except Exception as e:
        ret_code = -1
        stderr = str(e)

    callback(FinishedProcessInfo(ret_code, stdout, stderr))
    return None



from typing import Any, Dict

from pydantic import Field

from .base import RestBase


class ItemChannelLinkDefinition(RestBase):
    editable: bool
    channel_uid: str = Field(alias='channelUID')
    configuration: Dict[str, Any] = {}
    item_name: str = Field(alias='itemName')


class LinkNotFoundError(Exception):
    pass



from typing import Optional, List, Dict, Any

from .base_event import OpenhabEvent


class ThingStatusInfoEvent(OpenhabEvent):
    """
    :ivar str name:
    :ivar str status:
    :ivar str detail:
    """
    name: str
    status: str
    detail: str

    def __init__(self, name: str = '', status: str = '', detail: str = ''):
        super().__init__()

        self.name: str = name
        self.status: str = status
        self.detail: str = detail

    @classmethod
    def from_dict(cls, topic: str, payload: dict):
        # openhab/things/chromecast:chromecast:aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa/status
        return cls(name=topic[15:-7], status=payload['status'], detail=payload['statusDetail'])

    def __repr__(self):
        return f'<{self.__class__.__name__} name: {self.name}, status: {self.status}, detail: {self.detail}>'


class ThingStatusInfoChangedEvent(OpenhabEvent):
    """
    :ivar str name:
    :ivar str status:
    :ivar str detail:
    :ivar str old_status:
    :ivar str old_detail:
    """
    name: str
    status: str
    detail: str
    old_status: str
    old_detail: str

    def __init__(self, name: str = '', status: str = '', detail: str = '', old_status: str = '', old_detail: str = ''):
        super().__init__()

        self.name: str = name
        self.status: str = status
        self.detail: str = detail
        self.old_status: str = old_status
        self.old_detail: str = old_detail

    @classmethod
    def from_dict(cls, topic: str, payload: dict):
        # openhab/things/chromecast:chromecast:aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa/statuschanged
        name = topic[15:-14]
        new, old = payload
        return cls(
            name=name, status=new['status'], detail=new['statusDetail'],
            old_status=old['status'], old_detail=old['statusDetail']
        )

    def __repr__(self):
        return f'<{self.__class__.__name__} name: {self.name}, ' \
               f'status: {self.status}, detail: {self.detail}, ' \
               f'old_status: {self.old_status}, old_detail: {self.old_detail}>'


class ThingConfigStatusInfoEvent(OpenhabEvent):
    """
    :ivar str name:
    :ivar Dict[str, str] config_messages:
    """
    name: str
    config_messages: Dict[str, str]

    def __init__(self, name: str = '', config_messages: Optional[Dict[str, str]] = None):
        super().__init__()

        self.name: str = name
        self.config_messages: Dict[str, str] = config_messages if config_messages is not None else {}

    @classmethod
    def from_dict(cls, topic: str, payload: dict):
        # openhab/things/zwave:device:gehirn:node29/config/status
        name = topic[15:-14]
        msgs = payload['configStatusMessages']
        return cls(
            name=name, config_messages={param_name: msg_type for d in msgs for param_name, msg_type in d.items()}
        )

    def __repr__(self):
        return f'<{self.__class__.__name__} name: {self.name}, config_messages: {self.config_messages}>'


class ThingFirmwareStatusInfoEvent(OpenhabEvent):
    """
    :ivar str name:
    :ivar str status:
    """
    name: str
    status: str

    def __init__(self, name: str = '', status: str = ''):
        super().__init__()
        self.name: str = name
        self.status: str = status

    @classmethod
    def from_dict(cls, topic: str, payload: dict):
        # 'openhab/things/zwave:device:controller:my_node/firmware/status'
        return cls(name=topic[15:-16], status=payload['firmwareStatus'])

    def __repr__(self):
        return f'<{self.__class__.__name__} name: {self.name} status: {self.status}>'


class ThingRegistryBaseEvent(OpenhabEvent):
    """
    :ivar str name:
    :ivar str type:
    :ivar str label:
    :ivar List[Dict[str, Any]] channels:
    :ivar Dict[str, Any] configuration:
    :ivar Dict[str, str] properties:
    """
    name: str
    type: str
    label: str
    channels: List[Dict[str, Any]]
    configuration: Dict[str, Any]
    properties: Dict[str, str]

    def __init__(self, name: str = '', thing_type: str = '', label: str = '',
                 channels: Optional[List[Dict[str, Any]]] = None, configuration: Optional[Dict[str, Any]] = None,
                 properties: Optional[Dict[str, str]] = None):
        super().__init__()

        # use name instead of uuid
        self.name: str = name
        self.type: str = thing_type

        # optional entries
        self.label: str = label
        self.channels: List[Dict[str, Any]] = channels if channels is not None else []
        self.configuration: Dict[str, Any] = configuration if configuration is not None else {}
        self.properties: Dict[str, str] = properties if properties is not None else {}

    @classmethod
    def from_dict(cls, topic: str, payload: dict):
        # 'openhab/things/astro:sun:0a94363608/added'
        return cls(
            name=payload['UID'], thing_type=payload['thingTypeUID'], label=payload['label'],
            channels=payload.get('channels'), configuration=payload.get('configuration'),
            properties=payload.get('properties'),
        )

    def __repr__(self):
        return f'<{self.__class__.__name__} name: {self.name}>'


class ThingAddedEvent(ThingRegistryBaseEvent):
    pass


class ThingRemovedEvent(ThingRegistryBaseEvent):
    pass


class ThingUpdatedEvent(ThingRegistryBaseEvent):
    @classmethod
    def from_dict(cls, topic: str, payload: List[Dict[str, Any]]):

        payload = payload[0]
        return cls(
            name=payload['UID'], thing_type=payload['thingTypeUID'], label=payload['label'],
            channels=payload.get('channels'), configuration=payload.get('configuration'),
            properties=payload.get('properties'),
        )



import ast
from inspect import ismodule, isclass
from typing import List, Tuple, Callable, Any, Set, TypeVar

from immutables import Map
from stack_data import Variable

from HABApp.core.const.json import load_json, dump_json
from easyconfig.config_objs import ConfigObj
from .const import SEPARATOR_VARIABLES, PRE_INDENT


def _filter_expressions(name: str, value: Any) -> bool:
    # a is None = True
    if name.endswith(' is None'):
        return True

    # These types show no explicit types
    skipped_types = (type(None), str, float, int, list, dict, set, frozenset, Map)

    # type(b) = <class 'NoneType'>
    if name.startswith('type(') and value in skipped_types:
        return True

    # (str, int) = (<class 'str'>, <class 'int'>)
    if name.startswith('(') and name.endswith(')') and isinstance(value, tuple) and all(
            map(lambda x: x in skipped_types, value)):
        return True

    return False


SKIP_VARIABLE: Tuple[Callable[[str, Any], bool], ...] = (
    # module imports
    lambda name, value: ismodule(value),

    # type hints
    lambda name, value: name.startswith('typing.'),
    # type vars
    lambda name, value: isinstance(value, TypeVar),

    # functions
    lambda name, value: value is dump_json or value is load_json,

    # config value objs
    lambda name, value: isinstance(value, ConfigObj),

    # Expressions
    _filter_expressions
)

ORDER_VARIABLE: Tuple[Callable[[Variable], bool], ...] = (
    lambda x: isclass(x.value),
)


def skip_variable(var: Variable) -> bool:
    for func in SKIP_VARIABLE:
        name = var.name
        value = var.value
        if func(name, value):
            return True
    return False


def format_frame_variables(tb: List[str], stack_variables: List[Variable]):
    if not stack_variables:
        return None

    # remove variables that shall not be printed
    used_vars: List[Variable] = [v for v in stack_variables if not skip_variable(v)]

    # attributes
    dotted_names: Set[str] = {n.name.split('.')[0] for n in used_vars if '.' in n.name}

    # Sort output
    used_vars = sorted(used_vars, key=lambda x: (
        isinstance(x.nodes[0], ast.Compare),                                        # Compare objects last
        not any(map(
            lambda y: x.name == y or x.name.startswith(y + '.'), dotted_names)),    # Classes with attributes
        x.name.lower()                                                              # Name lowercase
    ))

    variables = {}

    # variables by order
    for add_var in ORDER_VARIABLE:
        for var in used_vars:
            name = var.name
            if name in variables:
                continue
            if add_var(var):
                variables[name] = var.value

    # rest of the variables
    for var in used_vars:
        if var.name not in variables:
            variables[var.name] = var.value

    if not variables:
        return None

    # Add variables to traceback
    tb.append(SEPARATOR_VARIABLES)

    for name, value in variables.items():
        tb.append(f'{" " * (PRE_INDENT + 1)}{name} = {repr(value)}')

    tb.append(SEPARATOR_VARIABLES)



from typing import Any

from HABApp.core.const import MISSING
from HABApp.core.events.filter.event import TypeBoundEventFilter
from HABApp.mqtt.events import MqttValueChangeEvent, MqttValueUpdateEvent


class MqttValueUpdateEventFilter(TypeBoundEventFilter):
    def __init__(self, value: Any = MISSING):
        super().__init__(MqttValueUpdateEvent, value=value)


class MqttValueChangeEventFilter(TypeBoundEventFilter):
    def __init__(self, value: Any = MISSING, old_value: Any = MISSING):
        super().__init__(MqttValueChangeEvent, value=value, old_value=old_value)



import logging
from typing import Dict, Set, Tuple, TYPE_CHECKING

from immutables import Map

import HABApp

from HABApp.core.internals import uses_item_registry
from HABApp.core.logger import log_warning

if TYPE_CHECKING:
    import HABApp.openhab.definitions.rest

log = logging.getLogger('HABApp.openhab.items')

Items = uses_item_registry()


def add_to_registry(item: 'HABApp.openhab.items.OpenhabItem', set_value=False):
    name = item.name
    for grp in item.groups:
        MEMBERS.setdefault(grp, set()).add(name)

    if not Items.item_exists(name):
        return Items.add_item(item)

    existing = Items.get_item(name)
    if isinstance(existing, item.__class__):
        # If we load directly through the API and not through an event we have to set the value
        if set_value:
            existing.set_value(item.value)

        # remove old groups
        for grp in set(existing.groups) - set(item.groups):
            MEMBERS.get(grp, set()).discard(name)

        # same type - it was only an item update (e.g. label)!
        existing.label    = item.label
        existing.tags     = item.tags
        existing.groups   = item.groups
        existing.metadata = item.metadata
        return None

    log_warning(log, f'Item type changed from {existing.__class__} to {item.__class__}')

    # Replace existing item with the updated definition
    Items.pop_item(name)
    Items.add_item(item)


def remove_from_registry(name: str):
    if not Items.item_exists(name):
        return None

    item = Items.get_item(name)  # type: HABApp.openhab.items.OpenhabItem
    for grp in item.groups:
        MEMBERS.get(grp, set()).discard(name)

    if isinstance(item, HABApp.openhab.items.GroupItem):
        MEMBERS.pop(name, None)

    Items.pop_item(name)
    return None


MEMBERS: Dict[str, Set[str]] = {}


def fresh_item_sync():
    MEMBERS.clear()


def get_members(group_name: str) -> Tuple['HABApp.openhab.items.OpenhabItem', ...]:
    ret = []
    for name in MEMBERS.get(group_name, []):
        item = Items.get_item(name)  # type: HABApp.openhab.items.OpenhabItem
        ret.append(item)
    return tuple(sorted(ret, key=lambda x: x.name))


# ----------------------------------------------------------------------------------------------------------------------
# Thing handling
# ----------------------------------------------------------------------------------------------------------------------
def add_thing_to_registry(thing: 'HABApp.openhab.definitions.rest.OpenhabThingDefinition'):
    name = thing.uid
    does_exist = Items.item_exists(name)

    # update existing
    if does_exist:
        existing = Items.get_item(name)  # type: HABApp.openhab.items.Thing
        if isinstance(existing, HABApp.openhab.items.Thing):
            existing.status = thing.status.status
            existing.status_detail = thing.status.detail
            return None

    # create new Thing
    new_thing = HABApp.openhab.items.Thing(name=name)
    new_thing.status        = thing.status.status
    new_thing.status_detail = thing.status.detail
    new_thing.label         = thing.label
    new_thing.configuration = Map(thing.configuration)
    new_thing.properties    = Map(thing.properties)

    if not does_exist:
        Items.add_item(new_thing)
        return None

    # Replace existing item with the updated definition
    log_warning(log, f'Item type changed from {existing.__class__} to {thing.__class__}')
    Items.pop_item(name)
    Items.add_item(new_thing)


def remove_thing_from_registry(name: str):
    if not Items.item_exists(name):
        return None
    return Items.pop_item(name)



from traceback import format_exception as _format_exception
from typing import Tuple, Union, Any, List

from stack_data import FrameInfo, Options

from .const import SEPARATOR_NEW_FRAME
from .format_frame import format_frame_info


def append_short_traceback(tb: List[str], e: Union[Exception, Tuple[Any, Any, Any]]):
    for line in _format_exception(*e) if isinstance(e, tuple) else _format_exception(type(e), e, e.__traceback__):
        for sub_lines in line.splitlines():
            tb.append(sub_lines.rstrip())


DEFAULT_OPTIONS = Options(include_signature=True, max_lines_per_piece=5)


def fallback_format(e: Exception, existing_traceback: List[str]) -> List[str]:
    # in case something goes wrong while formatting the traceback
    # we still want to show at least a small error message!
    new_tb = [f'Error while formatting traceback: {e}']
    append_short_traceback(new_tb, e)
    print(new_tb)

    # add traceback so we have some more information
    if existing_traceback:
        new_tb.append('')
        new_tb.append(SEPARATOR_NEW_FRAME)
        new_tb.append('Partial Traceback:')
        new_tb.append(SEPARATOR_NEW_FRAME)
        new_tb.extend(existing_traceback)
    return new_tb


def format_exception(e: Union[Exception, Tuple[Any, Any, Any]]) -> List[str]:
    tb: List[str] = []

    try:
        all_frames = tuple(FrameInfo.stack_data(e[2] if isinstance(e, tuple) else e.__traceback__, DEFAULT_OPTIONS))
        last_frame = len(all_frames) - 1

        added = True
        for i, frame_info in enumerate(all_frames):
            if isinstance(frame_info, FrameInfo):
                added = format_frame_info(tb, frame_info, is_last=i == last_frame)
            else:
                # repeated frames in case of recursion
                if added:
                    tb.append(f"... {frame_info.description} ...\n")

        # add a short traceback
        tb.append(SEPARATOR_NEW_FRAME)
        append_short_traceback(tb, e)

    except Exception as e:
        return fallback_format(e, tb)

    return tb



from typing import Any, FrozenSet, Optional

import HABApp.core
from .base_event import OpenhabEvent
from ..map_values import map_openhab_values


class ItemStateEvent(OpenhabEvent, HABApp.core.events.ValueUpdateEvent):
    """
    :ivar str name:
    :ivar value:
    """
    name: str
    value: Any

    def __init__(self, name: str = '', value: Any = None):
        super().__init__()

        # smarthome/items/NAME/state
        self.name: str = name
        self.value: Any = value

    @classmethod
    def from_dict(cls, topic: str, payload: dict):
        # smarthome/items/NAME/state
        return cls(topic[14:-6], map_openhab_values(payload['type'], payload['value']))

    def __repr__(self):
        return f'<{self.__class__.__name__} name: {self.name}, value: {self.value}>'


class ItemStateChangedEvent(OpenhabEvent, HABApp.core.events.ValueChangeEvent):
    """
    :ivar str name:
    :ivar value:
    :ivar old_value:
    """
    name: str
    value: Any
    old_value: Any

    def __init__(self, name: str = '', value: Any = None, old_value: Any = None):
        super().__init__()

        self.name: str = name
        self.value: Any = value
        self.old_value: Any = old_value

    @classmethod
    def from_dict(cls, topic: str, payload: dict):
        # smarthome/items/Ping/statechanged
        return cls(
            topic[14:-13],
            map_openhab_values(payload['type'], payload['value']),
            map_openhab_values(payload['oldType'], payload['oldValue'])
        )

    def __repr__(self):
        return f'<{self.__class__.__name__} name: {self.name}, value: {self.value}, old_value: {self.old_value}>'


class ItemCommandEvent(OpenhabEvent):
    """
    :ivar str name:
    :ivar value:
    """
    name: str
    value: Any

    def __init__(self, name: str = '', value: Any = None):
        super().__init__()

        self.name: str = name
        self.value: Any = value

    @classmethod
    def from_dict(cls, topic: str, payload: dict):
        # smarthome/items/NAME/command
        return cls(topic[14:-8], map_openhab_values(payload['type'], payload['value']))

    def __repr__(self):
        return f'<{self.__class__.__name__} name: {self.name}, value: {self.value}>'


class ItemAddedEvent(OpenhabEvent):
    """
    :ivar str name:
    :ivar str type:
    :ivar Optional[str] label:
    :ivar Tuple[str,...] tags:
    :ivar Tuple[str,...] group_names:
    """
    name: str
    type: str
    label: Optional[str]
    tags: FrozenSet[str]
    groups: FrozenSet[str]

    def __init__(self, name: str = '', type: str = '', label: Optional[str] = None,
                 tags: FrozenSet[str] = frozenset(), group_names: FrozenSet[str] = frozenset()):
        super().__init__()

        self.name: str = name
        self.type: str = type
        self.label: Optional[str] = label
        self.tags: FrozenSet[str] = tags
        self.groups: FrozenSet[str] = group_names

    @classmethod
    def from_dict(cls, topic: str, payload: dict):
        # {'topic': 'smarthome/items/NAME/added'
        # 'payload': '{"type":"Contact","name":"Test","tags":[],"groupNames":[]}'
        # 'type': 'ItemAddedEvent'}
        return cls(
            payload['name'], payload['type'], label=payload.get('label'),
            tags=frozenset(payload['tags']), group_names=frozenset(payload['groupNames'])
        )

    def __repr__(self):
        tags = f' {{{", ".join(sorted(self.tags))}}}' if self.tags else ""
        grps = f' {{{", ".join(sorted(self.groups))}}}' if self.groups else ""
        return f'<{self.__class__.__name__} name: {self.name}, type: {self.type}, tags:{tags}, groups:{grps}>'


class ItemUpdatedEvent(OpenhabEvent):
    """
    :ivar str name:
    :ivar str type:
    :ivar Tuple[str,...] tags:
    :ivar Tuple[str,...] group_names:
    """
    name: str
    type: str
    label: Optional[str]
    tags: FrozenSet[str]
    groups: FrozenSet[str]

    def __init__(self, name: str = '', type: str = '', label: Optional[str] = None,
                 tags: FrozenSet[str] = frozenset(), group_names: FrozenSet[str] = frozenset()):
        super().__init__()

        self.name: str = name
        self.type: str = type
        self.label: Optional[str] = label
        self.tags: FrozenSet[str] = tags
        self.groups: FrozenSet[str] = group_names

    @classmethod
    def from_dict(cls, topic: str, payload: dict):
        # smarthome/items/NAME/updated
        # 'payload': '[{"type":"Switch","name":"Test","tags":[],"groupNames":[]},
        #              {"type":"Contact","name":"Test","tags":[],"groupNames":[]}]',
        # 'type': 'ItemUpdatedEvent'
        new = payload[0]
        return cls(
            topic[14:-8], new['type'], label=new.get('label'),
            tags=frozenset(new['tags']), group_names=frozenset(new['groupNames'])
        )

    def __repr__(self):
        tags = f' {{{", ".join(sorted(self.tags))}}}' if self.tags else ""
        grps = f' {{{", ".join(sorted(self.groups))}}}' if self.groups else ""
        return f'<{self.__class__.__name__} name: {self.name}, type: {self.type}, tags:{tags}, groups:{grps}>'


class ItemRemovedEvent(OpenhabEvent):
    """
    :ivar str name:
    """
    name: str

    def __init__(self, name: str = ''):
        super().__init__()

        self.name = name

    @classmethod
    def from_dict(cls, topic: str, payload: dict):
        # smarthome/items/Test/removed
        return cls(topic[14:-8])

    def __repr__(self):
        return f'<{self.__class__.__name__} name: {self.name}>'


class ItemStatePredictedEvent(OpenhabEvent):
    """
    :ivar str name:
    :ivar value:
    """
    name: str
    value: Any

    def __init__(self, name: str = '', value: Any = None):
        super().__init__()

        # smarthome/items/NAME/state
        self.name: str = name
        self.value: Any = value

    @classmethod
    def from_dict(cls, topic: str, payload: dict):
        # 'smarthome/items/NAME/statepredicted'
        return cls(topic[14:-15], map_openhab_values(payload['predictedType'], payload['predictedValue']))

    def __repr__(self):
        return f'<{self.__class__.__name__} name: {self.name}, value: {self.value}>'


class GroupItemStateChangedEvent(OpenhabEvent):
    """
    :ivar str name:
    :ivar str item:
    :ivar value:
    :ivar old_value:
    """
    name: str
    item: str
    value: Any
    old_value: Any

    def __init__(self, name: str = '', item: str = '', value: Any = None, old_value: Any = None):
        super().__init__()

        self.name: str = name
        self.item: str = item

        self.value: Any = value
        self.old_value: Any = old_value

    @classmethod
    def from_dict(cls, topic: str, payload: dict):
        # 'smarthome/items/TestGroupAVG/TestNumber1/statechanged'
        parts = topic.split('/')

        return cls(
            parts[2], parts[3],
            map_openhab_values(payload['type'], payload['value']),
            map_openhab_values(payload['oldType'], payload['oldValue'])
        )

    def __repr__(self):
        return f'<{self.__class__.__name__} name: {self.name}, value: {self.value}, old_value: {self.old_value}>'



import io
import logging
from cProfile import Profile
from concurrent.futures import ThreadPoolExecutor
from pstats import SortKey
from pstats import Stats
from time import time
from typing import Callable, Any
from typing import Optional

from HABApp.core.internals import HINT_CONTEXT_OBJ
from HABApp.core.const import loop
from .base import WrappedFunctionBase, default_logger

WORKERS: Optional[ThreadPoolExecutor] = None


def create_thread_pool(count: int):
    global WORKERS
    assert isinstance(count, int) and count > 0

    default_logger.debug(f'Starting thread pool with {count:d} threads!')

    stop_thread_pool()
    WORKERS = ThreadPoolExecutor(count, 'HabAppWorker')


def stop_thread_pool():
    global WORKERS
    if WORKERS is not None:
        WORKERS.shutdown()
        WORKERS = None
        default_logger.debug('Thread pool stopped!')


async def run_in_thread_pool(func: Callable):
    return await loop.run_in_executor(
        WORKERS, func
    )


HINT_FUNC_SYNC = Callable[..., Any]


class WrappedThreadFunction(WrappedFunctionBase):

    def __init__(self, func: HINT_FUNC_SYNC,
                 warn_too_long=True,
                 name: Optional[str] = None,
                 logger: Optional[logging.Logger] = None,
                 context: Optional[HINT_CONTEXT_OBJ] = None):

        super(WrappedThreadFunction, self).__init__(name=name, func=func, logger=logger, context=context)
        assert callable(func)

        self.func = func

        self.warn_too_long: bool = warn_too_long
        self.time_submitted: float = 0.0

    def run(self, *args, **kwargs):
        self.time_submitted = time()
        WORKERS.submit(self.run_sync, *args, **kwargs)

    def run_sync(self, *args, **kwargs):
        start = time()

        # notify if we don't process quickly
        if start - self.time_submitted > 0.05:
            self.log.warning(f'Starting of {self.name} took too long: {start - self.time_submitted:.2f}s. '
                             f'Maybe there are not enough threads?')

        # start profiler
        pr = Profile()
        pr.enable()

        # Execute the function
        try:
            self.func(*args, **kwargs)
        except Exception as e:
            self.process_exception(e, *args, **kwargs)
            return None

        # disable profiler
        pr.disable()

        # log warning if execution takes too long
        duration = time() - start
        if self.warn_too_long and duration > 0.8:
            self.log.warning(f'Execution of {self.name} took too long: {duration:.2f}s')

            s = io.StringIO()
            ps = Stats(pr, stream=s).sort_stats(SortKey.CUMULATIVE)
            ps.print_stats(0.1)  # limit to output to 10% of the lines

            for line in s.getvalue().splitlines()[4:]:    # skip the amount of calls and "Ordered by:"
                if line:
                    self.log.warning(line)



from .definitions import ITEM_TYPES, ITEM_DIMENSIONS, GROUP_ITEM_FUNCTIONS
from .values import OnOffValue, PercentValue, UpDownValue, HSBValue, QuantityValue, OpenClosedValue, RawValue
from . import rest



from typing import Optional, FrozenSet, Mapping

from immutables import Map

from HABApp.openhab.items.base_item import OpenhabItem, MetaData
from ..definitions import QuantityValue


class NumberItem(OpenhabItem):
    """NumberItem which accepts and converts the data types from OpenHAB

    :ivar str name:
    :ivar Union[int, float] value:

    :ivar Optional[str] label:
    :ivar FrozenSet[str] tags:
    :ivar FrozenSet[str] groups:
    :ivar Mapping[str, MetaData] metadata:
    """


    @classmethod
    def from_oh(cls, name: str, value=None, label: Optional[str] = None, tags: FrozenSet[str] = frozenset(),
                groups: FrozenSet[str] = frozenset(), metadata: Mapping[str, MetaData] = Map()):
        if value is None:
            return cls(name, value, label=label, tags=tags, groups=groups, metadata=metadata)

        try:
            value = int(value)
        except ValueError:
            value = float(value)
        return cls(name, value, label, tags, groups, metadata)

    def set_value(self, new_value) -> bool:

        if isinstance(new_value, QuantityValue):
            return super().set_value(new_value.value)

        return super().set_value(new_value)



from typing import Any, Callable, Dict, Iterable, Optional, Union

from HABApp.core.internals import HINT_EVENT_FILTER_OBJ
from HABApp.core.items import BaseItem, HINT_ITEM_OBJ

from HABApp.core.lib.parameters import TH_POSITIVE_TIME_DIFF
from .listener_creator import ListenerCreatorBase, EventListenerCreator, \
    NoChangeEventListenerCreator, NoUpdateEventListenerCreator


class ListenerCreatorNotFoundError(Exception):
    pass


class EventListenerGroup:
    """Helper to create/cancel multiple event listeners simultaneously
    """

    def __init__(self):
        self._items: Dict[str, ListenerCreatorBase] = {}
        self._is_active = False

    @property
    def active(self) -> bool:
        """

        :return: True if the listeners are currently active
        """
        return self._is_active

    def listen(self):
        """Create all event listeners. If the event listeners are already active this will do nothing.
        """
        if self._is_active:
            return None
        self._is_active = True

        for o in self._items.values():
            o.listen()

    def cancel(self):
        """Cancel the active event listeners. If the event listeners are not active this will do nothing.
        """
        if not self._is_active:
            return None
        self._is_active = False

        for o in self._items.values():
            o.cancel()

    def activate_listener(self, name: str):
        """Resume a previously deactivated listener creator in the group.

        :param name: item name or alias of the listener
        :return: True if it was activated, False if it was already active
        """
        try:
            obj = self._items[name]
        except KeyError:
            raise ListenerCreatorNotFoundError(f'ListenerCreator for "{name}" not found!') from None
        if obj.active:
            return False

        obj.active = True
        if self._is_active:
            obj.listen()
        return True

    def deactivate_listener(self, name: str, cancel_if_active=True):
        """Exempt the listener creator from further listener/cancel calls

        :param name: item name or alias of the listener
        :param cancel_if_active: Cancel the listener if it is active
        :return: True if it was deactivated, False if it was already deactivated
        """
        try:
            obj = self._items[name]
        except KeyError:
            raise ListenerCreatorNotFoundError(f'ListenerCreator for "{name}" not found!') from None
        if not obj.active:
            return False

        if cancel_if_active:
            obj.cancel()
        obj.active = False
        return True

    def __add_objs(self, cls, item: Union[HINT_ITEM_OBJ, Iterable[HINT_ITEM_OBJ]], callback: Callable[[Any], Any],
                   arg, alias: Optional[str] = None):
        # alias -> single param
        if alias is not None:
            if not isinstance(item, BaseItem):
                raise ValueError('Only a single item can be passed together with alias')

        if isinstance(item, BaseItem):
            item = [item]

        for _item in item:
            name = _item.name if alias is None else alias
            self._items[name] = obj = cls(_item, callback, arg)
            if self._is_active:
                obj.listen()

    def add_listener(self, item: Union[HINT_ITEM_OBJ, Iterable[HINT_ITEM_OBJ]], callback: Callable[[Any], Any],
                     event_filter: HINT_EVENT_FILTER_OBJ, alias: Optional[str] = None) -> 'EventListenerGroup':
        """Add an event listener to the group

        :param item: Single or multiple items
        :param callback: Callback for the item(s)
        :param event_filter: Event filter for the item(s)
        :param alias: Alias if an item with the same name does already exist (e.g. if different callbacks shall be
                      created for the same item)
        :return: self
        """

        self.__add_objs(EventListenerCreator, item, callback, event_filter, alias)
        return self

    def add_no_update_watcher(self, item: Union[HINT_ITEM_OBJ, Iterable[HINT_ITEM_OBJ]], callback: Callable[[Any], Any],
                              seconds: TH_POSITIVE_TIME_DIFF, alias: Optional[str] = None
                              ) -> 'EventListenerGroup':
        """Add an no update watcher to the group. On ``listen`` this will create a no update watcher and
         the corresponding event listener that will trigger the callback

        :param item: Single or multiple items
        :param callback: Callback for the item(s)
        :param seconds: No update time for the no update watcher
        :param alias: Alias if an item with the same name does already exist (e.g. if different callbacks shall be
                      created for the same item)
        :return: self
        """
        self.__add_objs(NoUpdateEventListenerCreator, item, callback, seconds, alias)
        return self

    def add_no_change_watcher(self, item: Union[HINT_ITEM_OBJ, Iterable[HINT_ITEM_OBJ]], callback: Callable[[Any], Any],
                              seconds: TH_POSITIVE_TIME_DIFF, alias: Optional[str] = None
                              ) -> 'EventListenerGroup':
        """Add an no change watcher to the group. On ``listen`` this this will create a no change watcher and
         the corresponding event listener that will trigger the callback

        :param item: Single or multiple items
        :param callback: Callback for the item(s)
        :param seconds: No update time for the no change watcher
        :param alias: Alias if an item with the same name does already exist (e.g. if different callbacks shall be
                      created for the same item)
        :return: self
        """
        self.__add_objs(NoChangeEventListenerCreator, item, callback, seconds, alias)
        return self



import sys
from typing import TYPE_CHECKING, Any, Callable, Final


if TYPE_CHECKING:
    import HABApp
    import types
    import HABApp.rule_manager

_NAME: Final = '__HABAPP__HOOK__'


class HABAppRuleHook:

    @classmethod
    def in_dict(cls, obj: dict,
                cb_register_rule: Callable[['HABApp.rule.Rule'], Any],
                cb_suggest_name: Callable[['HABApp.rule.Rule'], str],
                runtime: 'HABApp.runtime.Runtime', rule_file: 'HABApp.rule_manager.RuleFile') -> dict:
        obj[_NAME] = cls(cb_register_rule, cb_suggest_name, runtime, rule_file)
        return obj

    def __init__(self,
                 cb_register_rule: Callable[['HABApp.rule.Rule'], Any],
                 cb_suggest_name: Callable[['HABApp.rule.Rule'], str],
                 runtime: 'HABApp.runtime.Runtime', rule_file: 'HABApp.rule_manager.RuleFile'):
        # callbacks
        self._cb_register: Final = cb_register_rule
        self._cb_suggest_name: Final = cb_suggest_name

        # runtime objs
        self.runtime: Final = runtime
        self.rule_file: Final = rule_file

    def register_rule(self, rule: 'HABApp.rule.Rule'):
        return self._cb_register(rule)

    def suggest_rule_name(self, rule: 'HABApp.rule.Rule') -> str:
        return self._cb_suggest_name(rule)


def get_rule_hook() -> HABAppRuleHook:

    depth = 0
    while True:
        depth += 1
        try:
            frame = sys._getframe(depth)    # type: types.FrameType
        except ValueError:
            raise RuntimeError('Rule files are not meant to be executed directly! '
                               'Put the file in the HABApp "rule" folder and HABApp will load it automatically.')

        _globals = frame.f_globals

        hook = _globals.get(_NAME, None)
        if hook is None:
            continue

        assert isinstance(hook, HABAppRuleHook)
        return hook



from datetime import datetime
from typing import Optional, FrozenSet, Mapping

from immutables import Map

from HABApp.openhab.items.base_item import OpenhabItem, MetaData


class DatetimeItem(OpenhabItem):
    """DateTimeItem which accepts and converts the data types from OpenHAB

    :ivar str name:
    :ivar datetime value:

    :ivar Optional[str] label:
    :ivar FrozenSet[str] tags:
    :ivar FrozenSet[str] groups:
    :ivar Mapping[str, MetaData] metadata:
    """

    @classmethod
    def from_oh(cls, name: str, value=None, label: Optional[str] = None, tags: FrozenSet[str] = frozenset(),
                groups: FrozenSet[str] = frozenset(), metadata: Mapping[str, MetaData] = Map()):
        if value is not None:
            dt = datetime.strptime(value, '%Y-%m-%dT%H:%M:%S.%f%z')
            # all datetime objs from openHAB have a timezone set so we can't easily compare them
            # --> TypeError: can't compare offset-naive and offset-aware datetime
            dt = dt.astimezone(tz=None)   # Changes datetime object so it uses system timezone
            value = dt.replace(tzinfo=None)  # Removes timezone awareness
        return cls(name, value, label=label, tags=tags, groups=groups, metadata=metadata)



from HABApp.openhab.items.base_item import OpenhabItem


class StringItem(OpenhabItem):
    """StringItem which accepts and converts the data types from OpenHAB

    :ivar str name:
    :ivar str value:

    :ivar Optional[str] label:
    :ivar FrozenSet[str] tags:
    :ivar FrozenSet[str] groups:
    :ivar Mapping[str, MetaData] metadata:
    """


class LocationItem(OpenhabItem):
    """LocationItem which accepts and converts the data types from OpenHAB

    :ivar str name:
    :ivar str value:

    :ivar Optional[str] label:
    :ivar FrozenSet[str] tags:
    :ivar FrozenSet[str] groups:
    :ivar Mapping[str, MetaData] metadata:
    """


class PlayerItem(OpenhabItem):
    """PlayerItem which accepts and converts the data types from OpenHAB

    :ivar str name:
    :ivar str value:

    :ivar Optional[str] label:
    :ivar FrozenSet[str] tags:
    :ivar FrozenSet[str] groups:
    :ivar Mapping[str, MetaData] metadata:
    """
